/*
 * Copyright (c) 1999 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * Copyright (c) 1999 Apple Computer, Inc.  All Rights Reserved.
 * The contents of this file constitute Original Code as defined in and are 
 * subject to the Apple Public Source License Version 1.1 (the "License").  
 * You may not use this file except in compliance with the License.  Please 
 * obtain a copy of the License at http://www.apple.com/publicsource and 
 * read it before using this file.
 * 
 * This Original Code and all software distributed under the License are 
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, EITHER 
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES, 
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY, FITNESS 
 * FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT.  Please see the License for 
 * the specific language governing rights and limitations under the 
 * License.
 * 
 * 
 * @APPLE_LICENSE_HEADER_END@
 */
/*
	File:		QTSSReflectorModule.cpp

	Contains:	Implementation of QTSSReflectorModule class. 
					
	
	$Log: QTSSReflectorModule.cpp,v $
	Revision 1.18  1999/12/21 01:43:25  lecroy
	*** empty log message ***
	
	Revision 1.17  1999/12/06 19:40:36  rtucker
	removed warning on second DECSCRIBE
	
	Revision 1.16  1999/12/04 09:29:06  serenyi
	Fix #2419499. Got rid of #DEBUG around code that wasn't just for debugging
	
	Revision 1.15  1999/11/15 01:24:32  serenyi
	Added in 4 char debugging codes
	
	Revision 1.14  1999/11/08 23:02:01  serenyi
	Updated to latest version of QTSS API
	
	Revision 1.13  1999/11/04 22:33:58  serenyi
	Added some error feedback to the relay parsing stuff
	
	Revision 1.12  1999/10/28 01:45:24  serenyi
	Fix 2404454 ReflectorSession & module were still assuming trackIDs were always 1-based
	
	Revision 1.11  1999/10/26 00:15:23  serenyi
	Make sure to check that there is no output for this session on a DESCRIBE
	
	Revision 1.10  1999/10/22 23:01:55  serenyi
	Made it so the reflector never reflects any more data after a TEARDOWN has been issued.
	
	Revision 1.9  1999/10/21 18:30:05  rtucker
	fixed "double send" bug in relfector
	
	Revision 1.8  1999/10/19 22:17:29  serenyi
	ReflectorSession now keeps track of its bitrate, each client puts that bitrate into the RTPSession
	
	Revision 1.7  1999/10/18 20:20:10  serenyi
	changed "dictType" to "objectType"
	
	Revision 1.6  1999/10/16 00:38:39  murata
	now using typed accessors for prefs dictionary.
	
	Revision 1.5  1999/10/14 01:17:27  rtucker
	updates for thinning
	
	Revision 1.4  1999/10/11 18:44:17  serenyi
	Use new RCFSourceInfo class for parsing inline relay configuratiion info
	
	Revision 1.3  1999/10/07 09:01:49  serenyi
	Changed bool to Bool16, changed QTSS_Dictionary to QTSS_Object
	
	Revision 1.2  1999/09/30 00:23:01  serenyi
	Changed ipAddr field of SourceInfo to be fDestIPAddr
	
	Revision 1.1.1.1  1999/09/22 23:01:07  lecroy
	importing
	
	Revision 1.3  1999/09/17 22:56:40  serenyi
	Fixed a memory leak: sdp data & local sdp data wasn't being destroyed
	
	Revision 1.2  1999/09/15 17:01:31  serenyi
	Made relay, reflector work
	
	Revision 1.1.1.1  1999/09/13 16:16:34  rtucker
	initial checkin from rrv14
	
	
	Revision 1.2  1999/02/19 23:08:25  ds
	Created
	

*/

#include "QTSSReflectorModule.h"
#include "QTSSModuleUtils.h"
#include "ReflectorSession.h"
#include "OSArrayObjectDeleter.h"
#include "QTSS_Private.h"

#include "OSMemory.h"
#include "OSRef.h"
#include "OSFileSource.h"
#include "IdleTask.h"
#include "Task.h"
#include "OS.h"
#include "Socket.h"
#include "SocketUtils.h"
#include "FilePrefsSource.h"
#include "OSArrayObjectDeleter.h"

//ReflectorOutput objects
#include "RTPSessionOutput.h"
#include "RelayOutput.h"

//SourceInfo objects
#include "SDPSourceInfo.h"
#include "RelaySDPSourceInfo.h"
#include "RCFSourceInfo.h"


#if DEBUG
#define REFLECTOR_MODULE_DEBUGGING 0
#else
#define REFLECTOR_MODULE_DEBUGGING 0
#endif

// ATTRIBUTES
static QTSS_AttributeID			sOutputAttr		 		= qtssIllegalAttrID;
static QTSS_AttributeID			sDelayPref 				= qtssIllegalAttrID;
static QTSS_AttributeID			sBucketSizePref 		= qtssIllegalAttrID;
static QTSS_AttributeID			sRelayPrefsPref 		= qtssIllegalAttrID;
static QTSS_AttributeID			sRelayStatsURLPref		= qtssIllegalAttrID;
static QTSS_AttributeID			sExpectedDigitFilenameErr=qtssIllegalAttrID;
static QTSS_AttributeID			sReflectorBadTrackIDErr = qtssIllegalAttrID;
static QTSS_AttributeID			sFoundNoSourcesErr 		= qtssIllegalAttrID;
static QTSS_AttributeID			sFoundNoDestsErr 		= qtssIllegalAttrID;
static QTSS_AttributeID			sFoundNoStreamsErr 		= qtssIllegalAttrID;
static QTSS_AttributeID			sFoundNoConfigFileErr 	= qtssIllegalAttrID;

// STATIC DATA

// ref to the prefs dictionary object
static QTSS_Object 		sServerPrefs;


static UInt32 				sDefaultDelayInMsec 		= 100;
static UInt32 				sDefaultBucketSize 			= 16;
static char* 				sDefaultRelayPrefs	 		= ""; // No default file location
static char* 				sDefaultRelayStatsURL		= ""; // No default URL

static UInt32 				sDelayInMsec 				= 100;
static UInt32 				sBucketSize					= 16;
static char* 				sRelayPrefs		 			= NULL;
static char* 				sRelayStatsURL	 			= NULL;
static StrPtrLen			sRequestHeader;

static OSRefTable* 	sSessionMap = NULL;

// Important strings
static StrPtrLen	sSDPSuffix(".sdp");
static StrPtrLen	sRCFSuffix(".rcf");
static StrPtrLen	sIncludeStr("Include");
static StrPtrLen	sRelaySourceStr("relay_source");

// FUNCTION PROTOTYPES

static QTSS_Error QTSSReflectorModuleDispatch(QTSS_Role inRole, QTSS_RoleParamPtr inParams);
static QTSS_Error 	Register(QTSS_Register_Params* inParams);
static QTSS_Error Initialize(QTSS_Initialize_Params* inParams);
static QTSS_Error Shutdown();
static QTSS_Error RereadPrefs();
static QTSS_Error ProcessRTSPRequest(QTSS_StandardRTSP_Params* inParams);
static QTSS_Error DoDescribe(QTSS_StandardRTSP_Params* inParams);
static char* 	GetSourceInfo(QTSS_RTSPRequestObject inRTSPRequest, SourceInfo** outInfo);
static ReflectorSession* 	FindOrCreateSession(SourceInfo* inInfo, QTSS_RTSPRequestObject inRequest);
static QTSS_Error DoSetup(QTSS_StandardRTSP_Params* inParams, ReflectorSession* inSession);
static QTSS_Error DoPlay(QTSS_StandardRTSP_Params* inParams, ReflectorSession* inSession);
static QTSS_Error DestroySession(QTSS_ClientSessionClosing_Params* inParams);
static void RemoveOutput(ReflectorOutput* inOutput, ReflectorSession* inSession);

// Relay Functions
static QTSS_Error Filter(QTSS_Filter_Params* inParams);
static void FindReflectorSessions(OSQueue* inSessionQueue);
static void RereadRelayPrefs();
static void FindRelayConfigFiles(OSQueue* inQueue);
static void ClearRelayConfigFiles(OSQueue* inQueue);
static void FindSourceInfos(OSQueue* inFileQueue, OSQueue* inSessionQueue);
static void ClearSourceInfos(OSQueue* inQueue);
static void EnQueuePath(OSQueue* inQueue, char* inFilePath);



// FUNCTION IMPLEMENTATIONS
#pragma mark __QTSS_REFLECTOR_MODULE__

QTSS_Error QTSSReflectorModule_Main(void* inPrivateArgs)
{
	return _stublibrary_main(inPrivateArgs, QTSSReflectorModuleDispatch);
}


QTSS_Error 	QTSSReflectorModuleDispatch(QTSS_Role inRole, QTSS_RoleParamPtr inParams)
{
	switch (inRole)
	{
		case QTSS_Register_Role:
			return Register(&inParams->regParams);
		case QTSS_Initialize_Role:
			return Initialize(&inParams->initParams);
		case QTSS_RereadPrefs_Role:
			return RereadPrefs();
		case QTSS_RTSPFilter_Role:
			return Filter(&inParams->rtspFilterParams);
		case QTSS_RTSPPreProcessor_Role:
			return ProcessRTSPRequest(&inParams->rtspRequestParams);
		case QTSS_ClientSessionClosing_Role:
			return DestroySession(&inParams->clientSessionClosingParams);
		case QTSS_Shutdown_Role:
			return Shutdown();
	}
	return QTSS_NoErr;
}


QTSS_Error Register(QTSS_Register_Params* inParams)
{
	// Do role & attribute setup
	(void)QTSS_AddRole(QTSS_Initialize_Role);
	(void)QTSS_AddRole(QTSS_Shutdown_Role);
	(void)QTSS_AddRole(QTSS_RTSPFilter_Role);
	(void)QTSS_AddRole(QTSS_RTSPPreProcessor_Role);
	(void)QTSS_AddRole(QTSS_ClientSessionClosing_Role);
	(void)QTSS_AddRole(QTSS_RereadPrefs_Role);
	
	// Add prefs attributes
	static char*		sDelayName 			= "reflector_delay";
	static char*		sBucketSizeName		= "reflector_bucket_size";
	static char*		sRelayPrefsName		= "relay_prefs_file";
	static char*		sRelayStatsURLName	= "relay_stats_url";

	(void)QTSS_AddAttribute(qtssPrefsObjectType, sDelayName, NULL);
	(void)QTSS_IDForAttr(qtssPrefsObjectType, sDelayName, &sDelayPref);
	
	(void)QTSS_AddAttribute(qtssPrefsObjectType, sBucketSizeName, NULL);
	(void)QTSS_IDForAttr(qtssPrefsObjectType, sBucketSizeName, &sBucketSizePref);

	(void)QTSS_AddAttribute(qtssPrefsObjectType, sRelayPrefsName, NULL);
	(void)QTSS_IDForAttr(qtssPrefsObjectType, sRelayPrefsName, &sRelayPrefsPref);

	(void)QTSS_AddAttribute(qtssPrefsObjectType, sRelayStatsURLName, NULL);
	(void)QTSS_IDForAttr(qtssPrefsObjectType, sRelayStatsURLName, &sRelayStatsURLPref);
	
	// Add text messages attributes
	static char*		sExpectedDigitFilenameName		= "QTSSReflectorModuleExpectedDigitFilename";
	static char*		sReflectorBadTrackIDErrName		= "QTSSReflectorModuleBadTrackID";
	static char*		sFoundNoSourcesErrName 			= "QTSSReflectorModuleNoRelaySources";
	static char*		sFoundNoDestsErrName 			= "QTSSReflectorModuleNoRelayDests";
	static char*		sFoundNoStreamsErrName 			= "QTSSReflectorModuleNoRelayStreams";
	static char*		sFoundNoConfigFileErrName 		= "QTSSReflectorModuleNoRelayConfig";

	(void)QTSS_AddAttribute(qtssTextMessagesObjectType, sExpectedDigitFilenameName, NULL);
	(void)QTSS_IDForAttr(qtssTextMessagesObjectType, sExpectedDigitFilenameName, &sExpectedDigitFilenameErr);

	(void)QTSS_AddAttribute(qtssTextMessagesObjectType, sReflectorBadTrackIDErrName, NULL);
	(void)QTSS_IDForAttr(qtssTextMessagesObjectType, sReflectorBadTrackIDErrName, &sReflectorBadTrackIDErr);
	
	(void)QTSS_AddAttribute(qtssTextMessagesObjectType, sFoundNoSourcesErrName, NULL);
	(void)QTSS_IDForAttr(qtssTextMessagesObjectType, sFoundNoSourcesErrName, &sFoundNoSourcesErr);
	
	(void)QTSS_AddAttribute(qtssTextMessagesObjectType, sFoundNoDestsErrName, NULL);
	(void)QTSS_IDForAttr(qtssTextMessagesObjectType, sFoundNoDestsErrName, &sFoundNoDestsErr);
	
	(void)QTSS_AddAttribute(qtssTextMessagesObjectType, sFoundNoStreamsErrName, NULL);
	(void)QTSS_IDForAttr(qtssTextMessagesObjectType, sFoundNoStreamsErrName, &sFoundNoStreamsErr);
	
	(void)QTSS_AddAttribute(qtssTextMessagesObjectType, sFoundNoConfigFileErrName, NULL);
	(void)QTSS_IDForAttr(qtssTextMessagesObjectType, sFoundNoConfigFileErrName, &sFoundNoConfigFileErr);
	
	// Add an RTP session attribute for tracking ReflectorSession objects
	static char*		sOutputName	= "QTSSReflectorModuleOutput";

	(void)QTSS_AddAttribute(qtssClientSessionObjectType, sOutputName, NULL);
	(void)QTSS_IDForAttr(qtssClientSessionObjectType, sOutputName, &sOutputAttr);
	
	// Reflector session needs to setup some parameters too.
	ReflectorSession::Register();
	// RTPSessionOutput needs to do the same
	RTPSessionOutput::Register();

	// Tell the server our name!
	static char* sModuleName = "QTSSReflectorModule";
	::strcpy(inParams->outModuleName, sModuleName);

	return QTSS_NoErr;
}


QTSS_Error Initialize(QTSS_Initialize_Params* inParams)
{
	// Setup module utils
	QTSSModuleUtils::Initialize(inParams->inMessages, inParams->inPrefs, inParams->inErrorLogStream);
	sSessionMap = new OSRefTable();
	RereadPrefs();

	sServerPrefs = inParams->inPrefs;

#if QTSS_REFLECTOR_EXTERNAL_MODULE
	// The reflector is dependent on a number of objects in the Common Utilities
	// library that get setup by the server if the reflector is internal to the
	// server.
	//
	// So, if the reflector is being built as a code fragment, it must initialize
	// those pieces itself
#if !MACOSXEVENTQUEUE
	::select_startevents();//initialize the select() implementation of the event queue
#endif
	OS::Initialize();
	Socket::Initialize();
	SocketUtils::Initialize();

	const UInt32 kNumReflectorThreads = 8;
	TaskThreadPool::AddThreads(kNumReflectorThreads);
	IdleTask::Initialize();
	Socket::StartThread();
#endif
	return QTSS_NoErr;
}

QTSS_Error Shutdown()
{
#if QTSS_REFLECTOR_EXTERNAL_MODULE
	TaskThreadPool::RemoveThreads();
#endif
	return QTSS_NoErr;
}

QTSS_Error RereadPrefs()
{
	delete [] sRelayPrefs;
	delete [] sRelayStatsURL;
	
	sDelayInMsec = QTSSModuleUtils::GetUInt32StringPref(sDelayPref, sDefaultDelayInMsec);
	sBucketSize = QTSSModuleUtils::GetUInt32StringPref(sBucketSizePref, sDefaultBucketSize);
	sRelayPrefs = QTSSModuleUtils::GetStringPref(sRelayPrefsPref, sDefaultRelayPrefs);
	sRelayStatsURL = QTSSModuleUtils::GetStringPref(sRelayStatsURLPref, sDefaultRelayStatsURL);

	RereadRelayPrefs();
	return QTSS_NoErr;
}


QTSS_Error ProcessRTSPRequest(QTSS_StandardRTSP_Params* inParams)
{
	QTSS_RTSPMethod* theMethod = NULL;
	UInt32 theLen = 0;
	if ((QTSS_GetValuePtr(inParams->inRTSPRequest, qtssRTSPReqMethod, 0,
			(void**)&theMethod, &theLen) != QTSS_NoErr) || (theLen != sizeof(QTSS_RTSPMethod)))
	{
		Assert(0);
		return QTSS_RequestFailed;
	}

	if (*theMethod == qtssDescribeMethod)
		return DoDescribe(inParams);
		
	RTPSessionOutput** theOutput = NULL;
	QTSS_Error theErr = QTSS_GetValuePtr(inParams->inClientSession, sOutputAttr, 0, (void**)&theOutput, &theLen);
	if ((theErr != QTSS_NoErr) || (theLen != sizeof(RTPSessionOutput*)))
		return QTSS_RequestFailed;
	
	switch (*theMethod)
	{
		case qtssSetupMethod:
			return DoSetup(inParams, (*theOutput)->GetReflectorSession());
		case qtssPlayMethod:
			return DoPlay(inParams, (*theOutput)->GetReflectorSession());
		case qtssTeardownMethod:
			// Tell the server that this session should be killed, and send a TEARDOWN response
			(void)QTSS_Teardown(inParams->inClientSession);
			(void)QTSS_SendStandardRTSPResponse(inParams->inRTSPRequest, inParams->inClientSession, 0);
			break;
		case qtssPauseMethod:
			(void)QTSS_Pause(inParams->inClientSession);
			(void)QTSS_SendStandardRTSPResponse(inParams->inRTSPRequest, inParams->inClientSession, 0);
			break;
		default:
			break;
	}			
	return QTSS_NoErr;
}


QTSS_Error DoDescribe(QTSS_StandardRTSP_Params* inParams)
{
	SourceInfo* theInfo = NULL;
	OSCharArrayDeleter theSDPData(GetSourceInfo(inParams->inRTSPRequest, &theInfo));
	if ((theInfo == NULL) || (!theInfo->IsReflectable()))
	{
		delete theInfo;
		return NULL;
	}
	
	ReflectorSession* theSession = FindOrCreateSession(theInfo, inParams->inRTSPRequest);
	if (theSession == NULL)
		return QTSS_RequestFailed;

	RTPSessionOutput** theOutput = NULL;
	UInt32 theLen = 0;
	QTSS_Error theErr = QTSS_GetValuePtr(inParams->inClientSession, sOutputAttr, 0, (void**)&theOutput, &theLen);
	//WarnV(theErr != QTSS_NoErr, "Found sOutputAttr on DESCRIBE"); // Can this ever happen?
	// yes the client will sometimes ask twice for a DESCRIBE

	// Only add a new RTPSessionOutput if there isn't already one for this RTP session
	if (theErr != QTSS_NoErr)
	{
		//ok, we've found or setup the proper reflector session, create an RTPSessionOutput object,
		//and add it to the session's list of outputs
		RTPSessionOutput* theNewOutput = new ('refO') RTPSessionOutput(inParams->inClientSession, theSession, sServerPrefs );
		theSession->AddOutput(theNewOutput);

		// And vice-versa, store this reflector session in the RTP session.
		(void)QTSS_SetValue(inParams->inClientSession, sOutputAttr, 0, &theNewOutput, sizeof(theNewOutput));
	}
		
	//above function has signalled that this request belongs to us, so let's respond
	iovec theDescribeVec[2] = { 0 };
	
	Assert(theSession->GetLocalSDP()->Ptr != NULL);
	theDescribeVec[1].iov_base = theSession->GetLocalSDP()->Ptr;
	theDescribeVec[1].iov_len = theSession->GetLocalSDP()->Len;
	QTSSModuleUtils::SendDescribeResponse(inParams->inRTSPRequest, inParams->inClientSession,
											&theDescribeVec[0], 2, theSession->GetLocalSDP()->Len);	
	delete theInfo;
	return QTSS_NoErr;
}

char*	GetSourceInfo(QTSS_RTSPRequestObject inRTSPRequest, SourceInfo** outInfo)
{
	// Return the SDP data as well.
	
	// Reflector needs to get a reflectable SourceInfo object in order to reflect.
	// Currently the only supported type for reflecting is SDPSourceInfo objects
	UInt32 thePathLen = 0;
	OSCharArrayDeleter sdpPath(QTSSModuleUtils::GetFullPath(inRTSPRequest, qtssRTSPReqFilePath,
																&thePathLen, &sSDPSuffix));

	StrPtrLen theFileData;

	// If the actual file path has a .sdp in it, first look for the URL without the extra .sdp
	if (thePathLen > (sSDPSuffix.Len * 2))
	{
		// Check and see if there is a .sdp in the file path
		StrPtrLen endOfPath(&sdpPath.GetObject()[thePathLen - (sSDPSuffix.Len * 2)], sSDPSuffix.Len);
		if (endOfPath.Equal(sSDPSuffix))
		{
			sdpPath.GetObject()[thePathLen - sSDPSuffix.Len] = '\0';
			OSFileSource::ReadEntireFile(sdpPath.GetObject(), &theFileData);
			if (theFileData.Len > 0)
				*outInfo = new ('SDPS') SDPSourceInfo(theFileData.Ptr, theFileData.Len);
		}
	}
	// If we haven't gotten an SDP file yet, try the URL with a .sdp appended.
	if (theFileData.Len == 0)
	{
		OSFileSource::ReadEntireFile(sdpPath.GetObject(), &theFileData);
		if (theFileData.Len > 0) // There is a SDP file!
			*outInfo = new SDPSourceInfo(theFileData.Ptr, theFileData.Len);
	}
	
	return theFileData.Ptr;
}

ReflectorSession* FindOrCreateSession(SourceInfo* inInfo, QTSS_RTSPRequestObject inRequest)
{
	//Build a sessionIDStr. Reflector sessions are uniquely identified by the (multicast)
	//source IP address and port #.
	char theSourceIDStr[SourceInfo::kSourceIDSize];
	StrPtrLen theSourceID(theSourceIDStr, SourceInfo::kSourceIDSize);
	inInfo->WriteSourceID(&theSourceID);
	
	QTSS_Error theErr = QTSS_NoErr;
	ReflectorSession* theSession = NULL;

	//Check to see if there is already a reflector session for this URL. If there is,
	//we are all set, just refcount it and return.
	{
		OSMutexLocker locker(sSessionMap->GetMutex());
		OSRef* theSessionRef = sSessionMap->Resolve(&theSourceID);
		
		if (theSessionRef == NULL)
		{
			//If this URL doesn't already have a reflector session, we must make a new
			//one. The first step is to parse the SDP file and see what addresses we
			//can find the multicast at.
			
			//create a reflector session, and bind the sockets
			theSession = new ('refs') ReflectorSession(inInfo, sDelayInMsec, sBucketSize);
			theErr = theSession->BindReflectorSockets(inRequest);
			if (theErr != QTSS_NoErr)
			{
				delete theSession;
				return NULL;
			}
			
			//put the session's ID into the session map.
			theErr = sSessionMap->Register(theSession->GetRef());
			Assert(theErr == QTSS_NoErr);

			//unless we do this, the refcount won't increment (and we'll delete the session prematurely
			OSRef* debug = sSessionMap->Resolve(&theSourceID);
			Assert(debug == theSession->GetRef());
		}
		else
			theSession = (ReflectorSession*)theSessionRef->GetObject();
			
		Assert(theSession != NULL);
		// Finally, if the ReflectorSession currently doesn't have localSDP, let's give it one
		theSession->UpdateLocalSDP(inInfo);
	}

	// We must have a valid ReflectorSession at this point	
	return theSession;
}


QTSS_Error DoSetup(QTSS_StandardRTSP_Params* inParams, ReflectorSession* inSession)
{
	//unless there is a digit at the end of this path (representing trackID), don't
	//even bother with the request
	StrPtrLen theDigit;
	(void)QTSS_GetValuePtr(inParams->inRTSPRequest, qtssRTSPReqFileDigit, 0, (void**)&theDigit.Ptr, &theDigit.Len);
	if (theDigit.Len == 0)
		return QTSSModuleUtils::SendErrorResponse(inParams->inRTSPRequest, qtssClientBadRequest,
													sExpectedDigitFilenameErr);
	UInt32 theTrackID = ::strtol(theDigit.Ptr, NULL, 10);
	QTSS_Error theErr = QTSS_NoErr;
	
	// Get info about this trackID
	StrPtrLen* thePayloadName = inSession->GetPayloadName(theTrackID);
	QTSS_RTPPayloadType thePayloadType = inSession->GetPayloadType(theTrackID);
	
	// If the PayloadName is NULL, we don't have a legit track, so return an error
	if (thePayloadName == NULL)
		return QTSSModuleUtils::SendErrorResponse(inParams->inRTSPRequest, qtssClientBadRequest,
													sReflectorBadTrackIDErr);
													
	QTSS_RTPStreamObject newStream = NULL;
	{
		//Because adding a stream
		//manipulates this RTPSession's stream queue, and because the reflector
		//uses that stream queue, we don't want the two running over each other.
		OSMutexLocker locker(inSession->GetMutex());
		theErr = QTSS_AddRTPStream(inParams->inClientSession, inParams->inRTSPRequest, &newStream, 0);
		if (theErr != QTSS_NoErr)
			return theErr;
	}
	
	// Set up dictionary items for this stream
	theErr = QTSS_SetValue(newStream, qtssRTPStrPayloadName, 0, thePayloadName->Ptr, thePayloadName->Len);
	Assert(theErr == QTSS_NoErr);
	theErr = QTSS_SetValue(newStream, qtssRTPStrPayloadType, 0, &thePayloadType, sizeof(thePayloadType));
	Assert(theErr == QTSS_NoErr);
	theErr = QTSS_SetValue(newStream, qtssRTPStrTrackID, 0, &theTrackID, sizeof(theTrackID));
	Assert(theErr == QTSS_NoErr);

	// Set the number of quality levels.
	static UInt32 sNumQualityLevels = ReflectorSession::kNumQualityLevels;
	theErr = QTSS_SetValue(newStream, qtssRTPStrNumQualityLevels, 0, &sNumQualityLevels, sizeof(sNumQualityLevels));
	Assert(theErr == QTSS_NoErr);
	
	//send the setup response
	(void)QTSS_SendStandardRTSPResponse(inParams->inRTSPRequest, newStream, 0);
	return QTSS_NoErr;
}


QTSS_Error DoPlay(QTSS_StandardRTSP_Params* inParams, ReflectorSession* inSession)
{
	// Tell the session what the bitrate of this reflection is. This is nice for logging,
	// it also allows the server to scale the TCP buffer size appropriately if we are
	// interleaving the data over TCP. This must be set before calling QTSS_Play so the
	// server can use it from within QTSS_Play
	UInt32 bitsPerSecond =	inSession->GetCurrentBitsPerSec();
	(void)QTSS_SetValue(inParams->inClientSession, qtssCliSesMovieAverageBitRate, 0, &bitsPerSecond, sizeof(bitsPerSecond));

	//Server shouldn't send RTCP (reflector does it), server shouldn't write rtp header either
	QTSS_Error theErr = QTSS_Play(inParams->inClientSession, inParams->inRTSPRequest, 0);
	if (theErr != QTSS_NoErr)
		return theErr;
	
	//and send a standard play response
	(void)QTSS_SendStandardRTSPResponse(inParams->inRTSPRequest, inParams->inClientSession, 0);
	return QTSS_NoErr;
}

QTSS_Error DestroySession(QTSS_ClientSessionClosing_Params* inParams)
{
	RTPSessionOutput** theOutput = NULL;
	UInt32 theLen = 0;
	QTSS_Error theErr = QTSS_GetValuePtr(inParams->inClientSession, sOutputAttr, 0, (void**)&theOutput, &theLen);
	if ((theErr != QTSS_NoErr) || (theLen != sizeof(RTPSessionOutput*)) || (theOutput == NULL))
		return QTSS_RequestFailed;
	
	RemoveOutput(*theOutput, (*theOutput)->GetReflectorSession());
	return QTSS_NoErr;
}

void RemoveOutput(ReflectorOutput* inOutput, ReflectorSession* inSession)
{
	// This function removes the output from the ReflectorSession, then
	// checks to see if the session should go away. If it should, this deletes it
	inSession->RemoveOutput(inOutput);

	//check if the ReflectorSession should be deleted
	//(it should if its ref count has dropped to 0)
	OSMutexLocker locker (sSessionMap->GetMutex());
	//decrement the ref count
	sSessionMap->Release(inSession->GetRef());
	if (inSession->GetRef()->GetRefCount() == 0)
	{
		Assert(inSession->GetNumOutputs() == 0);
		sSessionMap->UnRegister(inSession->GetRef());
		delete inSession;
	}	
	delete inOutput;
}

#pragma mark __RELAY_FUNCTIONS__

QTSS_Error Filter(QTSS_Filter_Params* inParams)
{
	UInt32 theLen = 0;
	char* theFullRequest = NULL;
	(void)QTSS_GetValuePtr(inParams->inRTSPRequest, qtssRTSPReqFullRequest, 0, (void**)&theFullRequest, &theLen);
	
	OSMutexLocker locker(RelayOutput::GetQueueMutex());

	// Check to see if this is a request we should handle
	if ((sRequestHeader.Ptr == NULL) || (sRequestHeader.Len == 0))
		return QTSS_NoErr;
	if ((theFullRequest == NULL) || (theLen < sRequestHeader.Len))
		return QTSS_NoErr;
	if (::memcmp(theFullRequest, sRequestHeader.Ptr, sRequestHeader.Len) != 0)
		return QTSS_NoErr;

	// Keep-alive should be off!
	Bool16 theFalse = false;
	(void)QTSS_SetValue(inParams->inRTSPRequest, qtssRTSPReqRespKeepAlive, 0, &theFalse, sizeof(theFalse));

	static StrPtrLen	sResponseHeader("HTTP/1.0 200 OK\r\nServer: QTSSReflectorModule/2.0\r\nConnection: Close\r\nContent-Type: text/html\r\n\r\n<HTML><TITLE>Relay Stats</TITLE><BODY>");
	(void)QTSS_Write(inParams->inRTSPRequest, sResponseHeader.Ptr, sResponseHeader.Len, NULL, 0);
	
	// First build up a queue of all ReflectorSessions with RelayOutputs. This way,
	// we can present a list that's sorted.
	OSQueue theSessionQueue;
	FindReflectorSessions(&theSessionQueue);
	
	static StrPtrLen	sNoRelays("<H2>There are no currently active relays</H2>");
	if (theSessionQueue.GetLength() == 0)
		(void)QTSS_Write(inParams->inRTSPRequest, sNoRelays.Ptr, sNoRelays.Len, NULL, 0);
	
	// Now go through our ReflectorSessionQueue, writing the info for each ReflectorSession,
	// and all the outputs associated with that session
	for (OSQueueIter iter(&theSessionQueue); !iter.IsDone(); iter.Next())
	{
		ReflectorSession* theSession = (ReflectorSession*)iter.GetCurrent()->GetEnclosingObject();
		(void)QTSS_Write(inParams->inRTSPRequest, theSession->GetSourceInfoHTML()->Ptr, theSession->GetSourceInfoHTML()->Len, NULL, 0);

		for (OSQueueIter iter2(RelayOutput::GetOutputQueue()); !iter2.IsDone(); iter2.Next())
		{
			RelayOutput* theOutput = (RelayOutput*)iter2.GetCurrent()->GetEnclosingObject();
			if (theSession == theOutput->GetReflectorSession())
			{
				(void)QTSS_Write(inParams->inRTSPRequest, theOutput->GetOutputInfoHTML()->Ptr, theOutput->GetOutputInfoHTML()->Len, NULL, 0);

				// Write current stats for this output
				char theStatsBuf[1024];
				::sprintf(theStatsBuf, "Current stats for this relay: %lu packets per second. %lu bits per second. %qu packets since it started. %qu bits since it started<P>", theOutput->GetCurPacketsPerSecond(), theOutput->GetCurBitsPerSecond(), theOutput->GetTotalPacketsSent(), theOutput->GetTotalBytesSent());
				(void)QTSS_Write(inParams->inRTSPRequest, &theStatsBuf[0], ::strlen(theStatsBuf), NULL, 0);
			}
		}
	}
	static StrPtrLen 	sResponseEnd("</BODY></HTML>");
	(void)QTSS_Write(inParams->inRTSPRequest, sResponseEnd.Ptr, sResponseEnd.Len, NULL, 0);
	
	for (OSQueueIter iter3(&theSessionQueue); !iter3.IsDone(); )
	{
		// Cleanup the memory we had allocated in FindReflectorSessions
		OSQueueElem* theElem = iter3.GetCurrent();
		iter3.Next();
		theSessionQueue.Remove(theElem);
		delete theElem;
	}
	return QTSS_NoErr;
}

void FindReflectorSessions(OSQueue* inSessionQueue)
{
	for (OSQueueIter theIter(RelayOutput::GetOutputQueue()); !theIter.IsDone(); theIter.Next())
	{
		RelayOutput* theOutput = (RelayOutput*)theIter.GetCurrent()->GetEnclosingObject();
		Bool16 found = false;
		
		// Check to see if we've already seen this ReflectorSession
		for (OSQueueIter theIter2(inSessionQueue); !theIter2.IsDone(); theIter2.Next())
		{
			if (theOutput->GetReflectorSession() == theIter2.GetCurrent()->GetEnclosingObject())
			{
				found = true;
				break;
			}
		}
		if (!found)
		{
			// We haven't seen this one yet, so put it on the queue.
			OSQueueElem* theElem = new ('fRFs') OSQueueElem(theOutput->GetReflectorSession());
			inSessionQueue->EnQueue(theElem);
		}
	}
}

void RereadRelayPrefs()
{
	// Find all the Relay configuration files
	OSQueue fileQueue;
	FindRelayConfigFiles(&fileQueue);
	
	// Now that we have the config files, go through each one,
	// constructing a SourceInfo object for each relayable thingy
	OSQueue sourceInfoQueue;
	FindSourceInfos(&fileQueue, &sourceInfoQueue);
	
	// Ok, we have all our SourceInfo objects. Now, lets alter the list of Relay
	// outputs to match
	
	// Go through the queue of Relay outputs, removing the source Infos that
	// are already going
	OSMutexLocker locker(RelayOutput::GetQueueMutex());

	// We must reread the relay stats URL here, because we have the RelayOutput
	// mutex so we know that Filter can't be filtering now
	delete [] sRequestHeader.Ptr;
	sRequestHeader.Ptr = NULL;
	sRequestHeader.Len = 0;

	if ((sRelayStatsURL != NULL) && (::strlen(sRelayStatsURL) > 0))
	{
		// sRequestHeader line should look like: GET /sRelayStatsURL HTTP
		sRequestHeader.Ptr = new ('RQHr') char[::strlen(sRelayStatsURL) + 15];
		::strcpy(sRequestHeader.Ptr, "GET /");
		::strcat(sRequestHeader.Ptr, sRelayStatsURL);
		::strcat(sRequestHeader.Ptr, " HTTP");
		sRequestHeader.Len = ::strlen(sRequestHeader.Ptr);
	}

	for (OSQueueIter iter(RelayOutput::GetOutputQueue()); !iter.IsDone(); )
	{
		RelayOutput* theOutput = (RelayOutput*)iter.GetCurrent()->GetEnclosingObject();
		
		Bool16 stillActive = false;
		
		// Check to see if this RelayOutput matches one of the streams in the
		// SourceInfo queue. If it does, this has 2 implications:
		//
		// 1) The stream in the SourceInfo that matches should NOT be setup as a
		// new RelayOutput, because it already exists. (Equal will set a flag in
		// the StreamInfo object saying as much)
		//
		// 2) This particular RelayOutput should remain (not be deleted).
		for (OSQueueIter iter2(&sourceInfoQueue); !iter2.IsDone(); iter2.Next())
		{
			if (theOutput->Equal((SourceInfo*)iter2.GetCurrent()->GetEnclosingObject()))
			{
				stillActive = true;
				break;
			}
		}
		iter.Next();	

		// This relay output is no longer on our list
		if (!stillActive)
			RemoveOutput(theOutput, theOutput->GetReflectorSession());
	}
	
	// We've pruned the list of RelayOutputs of all outputs that are no longer
	// going on. Now, all that is left to do is to create any new outputs (and sessions)
	// that are just starting up
	
	char theSourceIDStr[SourceInfo::kSourceIDSize];
	StrPtrLen theSourceID(theSourceIDStr, SourceInfo::kSourceIDSize);

	for (OSQueueIter iter3(&sourceInfoQueue); !iter3.IsDone(); iter3.Next())
	{
		SourceInfo* theInfo = (SourceInfo*)iter3.GetCurrent()->GetEnclosingObject();
		if (theInfo->GetNumNewOutputs() == 0) 	// Because we've already checked for Outputs
			continue;							// That already exist, we know which ones are new
			
		ReflectorSession* theSession = FindOrCreateSession(theInfo, NULL);
		if (theSession == NULL)
			continue;// if something very bad happened while trying to bind the source ports
					// (perhaps another process already has them or whatnot
#if DEBUG
		UInt32 resolveCount = 1;//we want to make sure we resolve for each new output
#endif

		// In order to Resolve this ReflectorSession, we'll need the StreamInfo ID
		theInfo->WriteSourceID(&theSourceID);

		for (UInt32 x = 0; x < theInfo->GetNumOutputs(); x++)
		{
			SourceInfo::OutputInfo* theOutputInfo = theInfo->GetOutputInfo(x);
			if (theOutputInfo->fAlreadySetup)
				continue;
				
			RelayOutput* theOutput = new ('rrot') RelayOutput(theInfo, x, theSession);
			theOutput->BindSocket();// ALERT!! WE ARE NOT CHECKING ERRORS HERE!!!
			theSession->AddOutput(theOutput);
			OSRef* debug = sSessionMap->Resolve(&theSourceID);
			Assert(debug->GetObject() == theSession);
#if DEBUG
			resolveCount++;
#endif
		}
		sSessionMap->Release(theSession->GetRef());// Release once right here, because we've
													// resolved once too many
		
#if DEBUG
		resolveCount--;
		Assert(resolveCount == theInfo->GetNumNewOutputs());
#endif
	}
	
	// We're done with the file queue now, free up the memory
	ClearRelayConfigFiles(&fileQueue);
	ClearSourceInfos(&sourceInfoQueue);
}

void FindRelayConfigFiles(OSQueue* inQueue)
{
	// This function goes through all the recursion in the relay config
	// files, and finds all of them. Each file path is put into a queue
	
	char* currentPath = sRelayPrefs;
	EnQueuePath(inQueue, sRelayPrefs);
	UInt32 numFiles = 0;
	
	for (OSQueueIter theIter(inQueue); !theIter.IsDone(); numFiles++)
	{
		// When we have parsed all the files on the list (and found no
		// additional files), thn we are done
		Assert(currentPath != NULL);
		FilePrefsSource theParser(currentPath, true);//Allow duplicate key names in the file
		
		// If we couldn't find the relay config file, or if the relay config file is
		// empty, log an error
		if ((numFiles == 0) && (theParser.GetNumKeys() == 0))
			QTSSModuleUtils::LogError(qtssWarningVerbosity, sFoundNoConfigFileErr, 0, sRelayPrefs);
			
		for (UInt32 x = 0; x < theParser.GetNumKeys(); x++)
		{
			StrPtrLen theKeyStr(theParser.GetKeyAtIndex(x));
			StrPtrLen theBodyStr(theParser.GetValueAtIndex(x));
			
			if ((theKeyStr.EqualIgnoreCase(sIncludeStr.Ptr, sIncludeStr.Len)) &&
				(theBodyStr.Len > sRCFSuffix.Len))
			{
				if (::memcmp(theBodyStr.Ptr + (theBodyStr.Len - sRCFSuffix.Len), sRCFSuffix.Ptr, sRCFSuffix.Len) == 0)
				{
					Bool16 found = false;
					
					// Make sure this path is not already in the list. If this happens,
					// and we don't detect it, we'd be looping here forever.
					for (OSQueueIter theIter2(inQueue); !theIter2.IsDone(); theIter2.Next())
					{
						if (theBodyStr.Equal((char*)theIter2.GetCurrent()->GetEnclosingObject()))
						{
							found = true;
							break;
						}
					}
					// If this path is not currently on the list, add it
					if (!found)
						EnQueuePath(inQueue, theBodyStr.Ptr);
				}
			}
		}
		
		// Move onto the next path. There maybe none left, in which case
		// we should break out of the loop above (currentPath just stays NULL)
		currentPath = NULL;
		theIter.Next();
		OSQueueElem* theNextFile = theIter.GetCurrent();
		if (theNextFile != NULL)
			currentPath = (char*)theNextFile->GetEnclosingObject();
	}
}

void ClearRelayConfigFiles(OSQueue* inQueue)
{
	for (OSQueueIter theIter(inQueue); !theIter.IsDone(); )
	{
		// Get the current queue element, and make sure to move onto the
		// next one, as the current one is going away.
		OSQueueElem* theElem = theIter.GetCurrent();
		theIter.Next();

		inQueue->Remove(theElem);
		char* thePath = (char*)theElem->GetEnclosingObject();
		delete [] thePath;
		delete theElem;
	}
}

void FindSourceInfos(OSQueue* inFileQueue, OSQueue* inSessionQueue)
{
	for (OSQueueIter theFileIter(inFileQueue); !theFileIter.IsDone(); theFileIter.Next())
	{
		FilePrefsSource theParser((char*)theFileIter.GetCurrent()->GetEnclosingObject(), true);
		Bool16 hasValidLine = false;
				
		// For this pass over the file, order matters, because groups of lines go together.
		// FilePrefsSource loads in all the prefs lines in reverse order, and we want to
		// iterate in top to bottom order, so that's why this loop is in reverse
		for (SInt32 x = theParser.GetNumKeys() - 1; x >= 0; x--)
		{
			StrPtrLen theKeyStr(theParser.GetKeyAtIndex(x));
			StrPtrLen theBodyStr(theParser.GetValueAtIndex(x));
			
			// Check to see if this line is an SDP include. If so, put it on our list
			if ((theKeyStr.EqualIgnoreCase(sIncludeStr.Ptr, sIncludeStr.Len)) &&
				(theBodyStr.Len > sSDPSuffix.Len))
			{
				if (::memcmp(theBodyStr.Ptr + (theBodyStr.Len - sSDPSuffix.Len), sSDPSuffix.Ptr, sSDPSuffix.Len) == 0)
				{
					hasValidLine = true;
					
					StrPtrLen theFileData;
					OSFileSource::ReadEntireFile(theBodyStr.Ptr, &theFileData);
					if (theFileData.Len > 0) // There is a Relay SDP file!
					{
						RelaySDPSourceInfo* theInfo = new ('RLsd') RelaySDPSourceInfo(&theFileData);

						// Only keep this sdp file around if there are input streams &
						// output streams described in it.
						if ((theInfo->GetNumStreams() == 0) || (theInfo->GetNumOutputs() == 0))
							delete theInfo;
						else
							inSessionQueue->EnQueue(theInfo->GetQueueElem());
						delete [] theFileData.Ptr;// We don't need the file data anymore
					}
				}
			}
			// If the line begins with "relay_source" it is the start of inlined
			// relay configuration information.
			else if (theKeyStr.EqualIgnoreCase(sRelaySourceStr.Ptr, sRelaySourceStr.Len))
			{
				hasValidLine = true;
				RCFSourceInfo* theRCFInfo = new ('RCFI') RCFSourceInfo(&theParser, x);
				
				if ((theRCFInfo->GetNumStreams() == 0) || (theRCFInfo->GetNumOutputs() == 0))
				{
					// Log an error to the error log
					if (theRCFInfo->GetNumStreams() == 0)
						QTSSModuleUtils::LogError(qtssWarningVerbosity, sFoundNoStreamsErr, 0);
					if (theRCFInfo->GetNumOutputs() == 0)
						QTSSModuleUtils::LogError(qtssWarningVerbosity, sFoundNoDestsErr, 0);
					delete theRCFInfo;
				}
				else
					inSessionQueue->EnQueue(theRCFInfo->GetQueueElem());
			}
		}
		
		// If we couldn't find anything useful in this file, log an error.
		if (!hasValidLine)
			QTSSModuleUtils::LogError(qtssWarningVerbosity, sFoundNoSourcesErr, 0, (char*)theFileIter.GetCurrent()->GetEnclosingObject());
	}
}

void ClearSourceInfos(OSQueue* inQueue)
{
	for (OSQueueIter theIter(inQueue); !theIter.IsDone(); )
	{
		// Get the current queue element, and make sure to move onto the
		// next one, as the current one is going away.
		OSQueueElem* theElem = theIter.GetCurrent();
		SourceInfo* theInfo = (SourceInfo*)theElem->GetEnclosingObject();

		theIter.Next();
	
		inQueue->Remove(theElem);
		delete theInfo;
	}
}


void EnQueuePath(OSQueue* inQueue, char* inFilePath)
{
	char* filePathCopy = new ('rcfP') char[::strlen(inFilePath) + 1];
	::strcpy(filePathCopy, inFilePath);
	
	OSQueueElem* newElem = new ('rcfQ') OSQueueElem(filePathCopy);
	inQueue->EnQueue(newElem);
}
