/*
 * Copyright (c) 1999 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * Copyright (c) 1999 Apple Computer, Inc.  All Rights Reserved.
 * The contents of this file constitute Original Code as defined in and are 
 * subject to the Apple Public Source License Version 1.1 (the "License").  
 * You may not use this file except in compliance with the License.  Please 
 * obtain a copy of the License at http://www.apple.com/publicsource and 
 * read it before using this file.
 * 
 * This Original Code and all software distributed under the License are 
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, EITHER 
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES, 
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY, FITNESS 
 * FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT.  Please see the License for 
 * the specific language governing rights and limitations under the 
 * License.
 * 
 * 
 * @APPLE_LICENSE_HEADER_END@
 */
/*
	File:		ReflectorSession.cpp

	Contains:	Implementation of object defined in ReflectorSession.h. 
					
	$Log: ReflectorSession.cpp,v $
	Revision 1.19  1999/12/23 18:46:19  rtucker
	mod sleep calc in ReflectorSession
	
	Revision 1.18  1999/12/21 01:43:26  lecroy
	*** empty log message ***
	
	Revision 1.17  1999/11/08 23:02:00  serenyi
	Updated to latest version of QTSS API
	
	Revision 1.16  1999/10/28 20:59:20  serenyi
	Got rid of dup packet WarnV, modified wakeup time assert
	
	Revision 1.15  1999/10/28 01:45:25  serenyi
	Fix 2404454 ReflectorSession & module were still assuming trackIDs were always 1-based
	
	Revision 1.14  1999/10/27 17:52:45  serenyi
	ntegrated Matthias's MacOS / GUSI changes
	
	Revision 1.13  1999/10/27 03:28:19  lecroy
	merged in changes from Akamai
	
	Revision 1.12  1999/10/26 23:57:02  rtucker
	enhanced reflector performance with large packet queues
	
	Revision 1.11  1999/10/25 16:25:08  rtucker
	;left a test switch off in the last check in
	
	Revision 1.10  1999/10/25 14:42:15  rtucker
	redeced warning count
	
	Revision 1.9  1999/10/21 18:30:06  rtucker
	fixed "double send" bug in relfector
	
	Revision 1.8  1999/10/19 22:17:29  serenyi
	ReflectorSession now keeps track of its bitrate, each client puts that bitrate into the RTPSession
	
	Revision 1.7  1999/10/18 20:20:10  serenyi
	changed "dictType" to "objectType"
	
	Revision 1.6  1999/10/16 00:57:28  rtucker
	fix reflector bugs from thinning
	
	Revision 1.5  1999/10/14 01:17:28  rtucker
	updates for thinning
	
	Revision 1.4  1999/10/11 18:43:38  serenyi
	Added some debugging info
	
	Revision 1.3  1999/10/07 09:01:49  serenyi
	Changed bool to Bool16, changed QTSS_Dictionary to QTSS_Object
	
	Revision 1.2  1999/09/30 00:23:02  serenyi
	Changed ipAddr field of SourceInfo to be fDestIPAddr
	
	Revision 1.1.1.1  1999/09/22 23:01:08  lecroy
	importing
	
	Revision 1.3  1999/09/17 22:56:41  serenyi
	Fixed a memory leak: sdp data & local sdp data wasn't being destroyed
	
	Revision 1.2  1999/09/15 17:01:31  serenyi
	Made relay, reflector work
	
	Revision 1.1.1.1  1999/09/13 16:16:34  rtucker
	initial checkin from rrv14
	
	
	Revision 1.1  1999/02/19 23:10:05  ds
	Created
	

*/

#include "ReflectorSession.h"
#include "RTCPPacket.h"
#include "SocketUtils.h"
#include "EventContext.h"

#include "OSMemory.h"
#include "OS.h"
#include "atomic.h"

#include "QTSSModuleUtils.h"

#include <sys/errno.h>


#if DEBUG
#define REFLECTOR_SESSION_DEBUGGING 0
#else
#define REFLECTOR_SESSION_DEBUGGING 0
#endif

// ATTRIBUTES

static QTSS_AttributeID 	sCantBindReflectorSocketErr = qtssIllegalAttrID;
static QTSS_AttributeID 	sCantJoinMulticastGroupErr 	= qtssIllegalAttrID;


#pragma mark _REFLECTOR_SESSION_

void ReflectorSession::Register()
{
	// Add text messages attributes
	static char*		sCantBindReflectorSocket= "QTSSReflectorModuleCantBindReflectorSocket";
	static char*		sCantJoinMulticastGroup	= "QTSSReflectorModuleCantJoinMulticastGroup";
	
	(void)QTSS_AddAttribute(qtssTextMessagesObjectType, sCantBindReflectorSocket, NULL);
	(void)QTSS_IDForAttr(qtssTextMessagesObjectType, sCantBindReflectorSocket, &sCantBindReflectorSocketErr);

	(void)QTSS_AddAttribute(qtssTextMessagesObjectType, sCantJoinMulticastGroup, NULL);
	(void)QTSS_IDForAttr(qtssTextMessagesObjectType, sCantJoinMulticastGroup, &sCantJoinMulticastGroupErr);
}

ReflectorSession::ReflectorSession(SourceInfo* inInfo, UInt32 inBucketDelayInMsec, UInt32 inBucketSize)
: 	fMutex('rarm'),
	fNumStreams(inInfo->GetNumStreams()),
	fSourceID(fSourceIDBuf, SourceInfo::kSourceIDSize),
 	fOutputArray(NULL),
 	fBucketSize(inBucketSize),
	fNumBuckets(kMinNumBuckets),
	fFirstEmptyBucket(1),
	fClientSpacing(inBucketDelayInMsec),
	fNumElements(0),
	fFormatter(fHTMLBuf, kMaxHTMLSize),
	fCurrentBitRate(0),
	fLastBitRateSample(OS::Milliseconds()), // don't calculate our first bit rate until kBitRateAvgIntervalInMilSecs has passed!
	fBytesSentInThisInterval(0)
{
	Assert(inInfo->GetNumStreams() > 0);
	
	// Store the SourceInfo unique ID
	inInfo->WriteSourceID(&fSourceID);
	fRef.Set(fSourceID, this);
	
	//allocate the session array
	this->AllocateSessionArray(fNumBuckets);
	
	// Begin writing our source description HTML (used by the relay)
	// Line looks like: Relay Source: 17.221.98.239, Ports: 5430 5432 5434
	// Note, we do nothing here to prevent overwriting our 255 byte buffer
	static StrPtrLen sHTMLStart("<H2>Relay Source: ");
	static StrPtrLen sPorts(", Ports: ");
	static StrPtrLen sHTMLEnd("</H2><BR>");
	
	// First, format the source addr as a dotted decimal string
	char theIPAddrBuf[20];
	StrPtrLen theIPAddr(theIPAddrBuf, 20);
	struct in_addr theAddr;
	theAddr.s_addr = htonl(inInfo->GetStreamInfo(0)->fDestIPAddr);
	SocketUtils::ConvertAddrToString(theAddr, &theIPAddr);
	
	// Begin writing the HTML
	fFormatter.Put(sHTMLStart);
	fFormatter.Put(theIPAddr);
	fFormatter.Put(sPorts);

	// Allocate all our ReflectorStreams, using the SourceInfo
	fStreamArray = new ('rfsa') ReflectorStream*[fNumStreams];
	for (UInt32 x = 0; x < fNumStreams; x++)
	{
		SourceInfo::StreamInfo* theStreamInfoP = inInfo->GetStreamInfo(x);
		fStreamArray[x] = new ('rfsm') ReflectorStream(this, theStreamInfoP);

		fFormatter.Put(theStreamInfoP->fPort);
		fFormatter.PutSpace();
	}
	fFormatter.Put(sHTMLEnd);

	// Setup the StrPtrLen to point to the right stuff
	fSourceInfoHTML.Ptr = fFormatter.GetBufPtr();
	fSourceInfoHTML.Len = fFormatter.GetCurrentOffset();
	
#if REFLECTOR_SESSION_DEBUGGING > 3
	fFormatter.PutTerminator();
	printf("Adding new ReflectorSession: %s\n", fSourceInfoHTML.Ptr);
	printf("Client spacing: %ld\n", inBucketDelayInMsec);
#endif
}


ReflectorSession::~ReflectorSession()
{
#if REFLECTOR_SESSION_DEBUGGING > 1
	printf("Removing ReflectorSession: %s\n", fSourceInfoHTML.Ptr);
#endif

	CloseSourceSockets();
	Assert(fNumElements == 0);

	for (UInt32 x = 0; x < fNumStreams; x++)
		delete fStreamArray[x];
	delete [] fStreamArray;

	//delete every client Bucket
	for (UInt32 y = 0; y < fNumBuckets; y++)
		delete [] fOutputArray[y];
	delete [] fOutputArray;
	
	// If there is local SDP, delete it.
	delete [] fLocalSDP.Ptr;
}

void	ReflectorSession::UpdateLocalSDP(SourceInfo* inInfo)
{
	if (fLocalSDP.Ptr == NULL)
		fLocalSDP.Ptr = inInfo->GetLocalSDP(&fLocalSDP.Len);
}


/*****************************************************
/	ReflectorSession::AddOutput
/
/	Add and output to the first available bucket in this ReflectorSession
/
/	Typically this occurs in respoinse to the DESCRIBE RTSP request.
/
/
*/

void ReflectorSession::AddOutput(ReflectorOutput* inOutput)
{
	OSMutexLocker locker(&fMutex);
	if (fNumElements == (fBucketSize * fNumBuckets))
		this->AllocateSessionArray(fNumBuckets * 2);

	Assert(fNumElements < fBucketSize * fNumBuckets);

	//find the first open spot in the array
	for (UInt32 x = 0; x < fNumBuckets; x++)
	{
		for(UInt32 y = 0; y < fBucketSize; y++)
		{
			if (fOutputArray[x][y] == NULL)
			{
				fOutputArray[x][y] = inOutput;
#if REFLECTOR_SESSION_DEBUGGING > 1
				printf("Adding new output (0x%lx) to bucket %ld, index %ld,\nnum buckets %li bucketSize: %li \n",(long)inOutput, x, y, (long)fNumBuckets, (long)fBucketSize);
#endif
				//if this session is filling a bucket that was
				//previously empty, make sure to track that!
				if (x >= fFirstEmptyBucket)
					fFirstEmptyBucket = x+1;
				fNumElements++;
				return;
			}
		}
	}
	Assert(0);
}

void  ReflectorSession::RemoveOutput(ReflectorOutput* inOutput)
{
	OSMutexLocker locker(&fMutex);
	Assert(fNumElements > 0);
	
	//look at all the indexes in the array
	for (UInt32 x = 0; x < fFirstEmptyBucket; x++)
	{
		for (UInt32 y = 0; y < fBucketSize; y++)
		{
			//The array may have blank spaces!
			if (fOutputArray[x][y] == inOutput)
			{
				fOutputArray[x][y] = NULL;//just clear out the pointer
				
				//check to see if eliminating this session has created a new empty
				//bucket. This is important to track because it eases the
				//buffering / smoothing burden
				if (fFirstEmptyBucket == (x+1))
				{
					Bool16 isEmpty = true;
					for (UInt32 z = 0; z < fBucketSize; z++)
						if (fOutputArray[x][z] != NULL)
							isEmpty = false;
					if (isEmpty)
						fFirstEmptyBucket--;
				}
				
#if REFLECTOR_SESSION_DEBUGGING > 1
				printf("Removing output from bucket %ld, index %ld\n",x,y);
#endif
				fNumElements--;
				return;				
			}
		}
	}
	Assert(0);
}

void ReflectorSession::AllocateSessionArray(UInt32 inNumBuckets)
{
	Bucket* oldArray = fOutputArray;
	//allocate the 2-dimensional array
	fOutputArray = new ('buck') Bucket[inNumBuckets];
	for (UInt32 x = 0; x < inNumBuckets; x++)
	{
		fOutputArray[x] = new ('rsma') ReflectorOutput*[fBucketSize];
		::memset(fOutputArray[x], 0, sizeof(ReflectorOutput*) * fBucketSize);
	}
	
	//copy over the old information if there was an old array
	if (oldArray != NULL)
	{
		Assert(inNumBuckets > fNumBuckets);
		for (UInt32 y = 0; y < fNumBuckets; y++)
		{
			::memcpy(fOutputArray[y],oldArray[y], fBucketSize * sizeof(ReflectorOutput*));
			delete [] oldArray[y];
		}
		delete [] oldArray;
	}
	fNumBuckets = inNumBuckets;
}

QTSS_Error ReflectorSession::BindReflectorSockets(QTSS_RTSPRequestObject inRequest)
{
	for (UInt32 x = 0; x < fNumStreams; x++)
	{
		QTSS_Error theError = fStreamArray[x]->BindSockets(inRequest);
		if (theError != QTSS_NoErr)
			return theError;
	}
	return QTSS_NoErr;
}

void	ReflectorSession::CloseSourceSockets()
{
	for (UInt32 x = 0; x < fNumStreams; x++)
		fStreamArray[x]->CloseSockets();
}

Bool16 ReflectorSession::Equal(SourceInfo* inInfo)
{
	if ((fNumStreams == inInfo->GetNumStreams()) &&
		(fStreamArray[0]->fStreamInfo.fDestIPAddr == inInfo->GetStreamInfo(0)->fDestIPAddr) &&
		(fStreamArray[0]->fStreamInfo.fPort == inInfo->GetStreamInfo(0)->fPort))
		return true;
	return false;
}

StrPtrLen* ReflectorSession::GetPayloadName(UInt32 trackID)
{
	for (UInt32 x = 0; x < fNumStreams; x++)
	{
		if (fStreamArray[x]->fStreamInfo.fTrackID == trackID)
			return fStreamArray[x]->GetPayloadName();
	}
	return NULL;
}

QTSS_RTPPayloadType ReflectorSession::GetPayloadType(UInt32 trackID)
{
	for (UInt32 x = 0; x < fNumStreams; x++)
	{
		if (fStreamArray[x]->fStreamInfo.fTrackID == trackID)
			return fStreamArray[x]->GetPayloadType();
	}
	return qtssUnknownPayloadType;
}


#pragma mark __REFLECTOR_STREAM__

ReflectorSocketPool	ReflectorStream::sSocketPool;

ReflectorStream::ReflectorStream(ReflectorSession* inSession, SourceInfo::StreamInfo* inInfo)
: 	fSockets(NULL),
	fSession(inSession),
	fRTPSender(this, inSession, qtssWriteFlagsIsRTP),
	fRTCPSender(this, inSession, qtssWriteFlagsIsRTCP),
	fStreamInfo(*inInfo),//copy construct the streamInfo
	fDestRTCPAddr(0),
	fDestRTCPPort(0)
{
	//copy the codec name to a new, local buffer
	if ((inInfo->fPayloadName.Ptr != NULL) && (inInfo->fPayloadName.Len > 0))
	{
		fStreamInfo.fPayloadName.Ptr = new ('CODE') char[inInfo->fPayloadName.Len + 1];
		::memcpy(fStreamInfo.fPayloadName.Ptr, inInfo->fPayloadName.Ptr, inInfo->fPayloadName.Len);
		fStreamInfo.fPayloadName.Ptr[inInfo->fPayloadName.Len] = '\0';
	}
	
	//write as much of the RTCP RR as is possible right now (most of it never changes)
	UInt32 theSsrc = (UInt32)::rand();
	char theTempCName[RTCPSRPacket::kMaxCNameLen];
	UInt32 cNameLen = RTCPSRPacket::GetACName(theTempCName);
	
	//write the RR (just header + ssrc)
	UInt32* theRRWriter = (UInt32*)&fReceiverReportBuffer[0];
	*theRRWriter = htonl(0x80c90001);
	theRRWriter++;
	*theRRWriter = htonl(theSsrc);
	theRRWriter++;

	//SDES length is the length of the CName, plus 2 32bit words, minus 1
	*theRRWriter = htonl(0x81ca0000 + (cNameLen >> 2) + 1);
	theRRWriter++;
	*theRRWriter = htonl(theSsrc);
	theRRWriter++;
	::memcpy(theRRWriter, theTempCName, cNameLen);
	theRRWriter += cNameLen >> 2;
	
	//APP packet format, QTSS specific stuff
	*theRRWriter = htonl(0x80cc0008);
	theRRWriter++;
	*theRRWriter = htonl(theSsrc);
	theRRWriter++;
	*theRRWriter = 'QTSS';
	theRRWriter++;
	*theRRWriter = htonl(0);
	theRRWriter++;
	*theRRWriter = htonl(0x00000004);
	theRRWriter++;
	*theRRWriter = htonl(0x6579000c);
	theRRWriter++;
	
	fEyeLocation = theRRWriter;
	fReceiverReportSize = kReceiverReportSize + kAppSize + cNameLen;
	
	// If the source is a multicast, we should send our receiver reports
	// to the multicast address
	if (SocketUtils::IsMulticastIPAddr(fStreamInfo.fDestIPAddr))
	{
		fDestRTCPAddr = fStreamInfo.fDestIPAddr;
		fDestRTCPPort = fStreamInfo.fPort + 1;
	}
}

QTSS_Error ReflectorStream::BindSockets(QTSS_RTSPRequestObject inRequest)
{
	//get a pair of sockets
	fSockets = sSocketPool.GetUDPSocketPair(INADDR_ANY, fStreamInfo.fPort, fStreamInfo.fDestIPAddr, fStreamInfo.fPort);
	if (fSockets == NULL)
		return QTSSModuleUtils::SendErrorResponse(inRequest, qtssServerInternal,
													sCantBindReflectorSocketErr);

#if !OS_SUPPORTS_UDP_DEST_ADDR
	//In order to have multiple multicasts on the same port pair, we
	//need to be able to demux based on destination addr, which is not
	//always supported by sockets implementations
 	if (((ReflectorSocket*)fSockets->GetSocketA())->HasSender())
		return QTSSModuleUtils::SendErrorResponse(inRequest, qtssServerInternal,
													sCantBindReflectorSocketErr);
#endif 		
	
	//If the broadcaster is sending RTP directly to us, we don't
	//need to join a multicast group because we're not using multicast
	if (SocketUtils::IsMulticastIPAddr(fStreamInfo.fDestIPAddr))
	{
		QTSS_Error err = fSockets->GetSocketA()->JoinMulticast(fStreamInfo.fDestIPAddr);
		if (err == QTSS_NoErr)
			err = fSockets->GetSocketB()->JoinMulticast(fStreamInfo.fDestIPAddr);
		if (err == QTSS_NoErr)
			err = fSockets->GetSocketA()->SetTtl(fStreamInfo.fTimeToLive);
		if (err == QTSS_NoErr)
			err = fSockets->GetSocketB()->SetTtl(fStreamInfo.fTimeToLive);
		if (err != QTSS_NoErr)
			return QTSSModuleUtils::SendErrorResponse(inRequest, qtssServerInternal,
														sCantJoinMulticastGroupErr);
	}

	//also put this stream onto the socket's queue of streams
	((ReflectorSocket*)fSockets->GetSocketA())->AddSender(&fRTPSender);
	((ReflectorSocket*)fSockets->GetSocketB())->AddSender(&fRTCPSender);

	//finally, register these sockets for events
	fSockets->GetSocketA()->RequestEvent(EV_RE);
	fSockets->GetSocketB()->RequestEvent(EV_RE);
	
	return QTSS_NoErr;
}

void ReflectorStream::CloseSockets()
{
	if (fSockets != NULL)
	{
		//first things first, let's take this stream off the socket's queue
		//of streams. This will basically ensure that no reflecting activity
		//can happen on this stream.
		((ReflectorSocket*)fSockets->GetSocketA())->RemoveSender(&fRTPSender);
		((ReflectorSocket*)fSockets->GetSocketB())->RemoveSender(&fRTCPSender);
		
		//leave the multicast group. Because this socket is shared amongst several
		//potential multicasts, we don't want to remain a member of a stale multicast
		if (SocketUtils::IsMulticastIPAddr(fStreamInfo.fDestIPAddr))
		{
			fSockets->GetSocketA()->LeaveMulticast(fStreamInfo.fDestIPAddr);
			fSockets->GetSocketB()->LeaveMulticast(fStreamInfo.fDestIPAddr);
		}
		//now release the socket pair
		sSocketPool.ReleaseUDPSocketPair(fSockets);
	}
	fSockets = NULL;
}

void ReflectorStream::SendReceiverReport()
{
	// Check to see if our destination RTCP addr & port are setup. They may
	// not be if the source is unicast and we haven't gotten any incoming packets yet
	if (fDestRTCPAddr == 0)
		return;
	
	UInt32 theEyeCount = fSession->fNumElements;
	
	//write the current eye count information based on the array
	//this is not really pre-emptive safe, but who cares... it's just statistics
	UInt32* theEyeWriter = fEyeLocation;
	*theEyeWriter = htonl(theEyeCount) & 0x7fffffff;//no idea why we do this!
	theEyeWriter++;
	*theEyeWriter = htonl(theEyeCount) & 0x7fffffff;
	theEyeWriter++;
	*theEyeWriter = htonl(0) & 0x7fffffff;
	
	//send the packet to the multicast RTCP addr & port for this stream
	(void)fSockets->GetSocketB()->SendTo(fDestRTCPAddr, fDestRTCPPort, fReceiverReportBuffer, fReceiverReportSize);
}


ReflectorSender::ReflectorSender(ReflectorStream* inStream, ReflectorSession* inSession, UInt32 inWriteFlag)
: 	fStream(inStream),
	fSession(inSession),
	fWriteFlag(inWriteFlag),
	fSocketQueueElem(this),
	fFirstNewPacketInQueue(NULL),
	fHasNewPackets(false),
	fNextTimeToRun(0),
	fLastRRTime(0),
	fLastQueueCleanupTime(0)
{}

ReflectorSender::~ReflectorSender()
{
	//dequeue and delete every buffer
	while (fPacketQueue.GetLength() > 0)
	{
		ReflectorPacket* packet = (ReflectorPacket*)fPacketQueue.DeQueue()->GetEnclosingObject();
		delete packet;
	}
}


Bool16 ReflectorSender::ShouldReflectNow(const SInt64& inCurrentTime, SInt64* ioWakeupTime)
{
	Assert(ioWakeupTime != NULL);
	//check to make sure there actually is work to do for this stream.
	if ((!fHasNewPackets) && ((fNextTimeToRun == 0) || (inCurrentTime < fNextTimeToRun)))
	{
		//We don't need to do work right now, but
		//this stream must still communicate when it needs to be woken up next
		SInt64 theWakeupTime = fNextTimeToRun - inCurrentTime;
		if ((fNextTimeToRun > 0) && (theWakeupTime < *ioWakeupTime))
			*ioWakeupTime = theWakeupTime;
#if !OS_SUPPORTS_UDP_DEST_ADDR
		//if this is the case, only one sender per socket is allowed, and the 1
		//sender should ALWAYS have work to do!
#if DEBUG
		if ((!fHasDroppedDupPacket) && (fWriteFlag == qtssWriteFlagsIsRTP))
			Assert(0);
#endif
#endif
		return false;
	}
	return true;	
}



#if REFLECTOR_SESSION_DEBUGGING
static UInt16 DGetPacketSeqNumber(StrPtrLen* inPacket)
{
	if (inPacket->Len < 4)
		return 0;
	
	//The RTP seq number is the second short of the packet
	UInt16* seqNumPtr = (UInt16*)inPacket->Ptr;
	return ntohs(seqNumPtr[1]);
}



#endif

/***********************************************************************************************
/	ReflectorSender::ReflectPackets
/	
/	There are n ReflectorSender's for n output streams per presentation.
/	
/	Each sender is associated with an array of ReflectorOutput's.  Each
/	output represents a client connection.  Each output has # RTPStream's. 
/	
/	When we write a packet to the ReflectorOutput he matches it's payload
/	to one of his streams and sends it there.
/	
/	To smooth the bandwitdth (server, not user) requirements of the reflected streams, the Sender
/	groups the ReflectorOutput's into buckets.  The input streams are reflected to
/	each bucket progressively later in time.  So rather than send a single packet
/	to say 1000 clients all at once, we send it to just the first 16, then then next 16 
/	100 ms later and so on.
/
/
/	intputs 	ioWakeupTime - relative time to call us again in MSec
/				inFreeQueue - queue of free packets.
*/


void ReflectorSender::ReflectPackets(SInt64* ioWakeupTime, OSQueue* inFreeQueue)
{
#if DEBUG
	Assert(ioWakeupTime != NULL);
#endif

#if REFLECTOR_SESSION_DEBUGGING 
	Bool16	printQueueLenOnExit = false;
#endif
	
	Bool16	isTimeToCleanUpQueue = false;
	SInt64 	currentTime = OS::Milliseconds();

	//make sure to reset these state variables
	fHasNewPackets = false;	
	fNextTimeToRun = -1;	// init to never
	
	//set flag for periodical tasks for this queue
	if ( currentTime > (fLastQueueCleanupTime + kCleanUpInterval) )
	{
		fLastQueueCleanupTime = currentTime;
		
		isTimeToCleanUpQueue = true;
		
		#if REFLECTOR_SESSION_DEBUGGING 
		printQueueLenOnExit = true;
		printf( "fPacketQueue len %li\n", (long)fPacketQueue.GetLength() );
		#endif	
	}
	
	//determine if we need to send a receiver report to the multicast source
	if ((fWriteFlag == qtssWriteFlagsIsRTCP) && (currentTime > (fLastRRTime + kRRInterval)))
	{
		fLastRRTime = currentTime;
		fStream->SendReceiverReport();
	}
	
	//the rest of this function must be atomic wrt the ReflectorSession, because
	//it involves iterating through the RTPSession array, which isn't thread safe
	OSMutexLocker locker(&fSession->fMutex);
	
	// Check to see if we should update the session's bitrate average
	if ((fSession->fLastBitRateSample + ReflectorSession::kBitRateAvgIntervalInMilSecs) < currentTime)
	{
		unsigned int intervalBytes = fSession->fBytesSentInThisInterval;
		(void)atomic_sub(&fSession->fBytesSentInThisInterval, intervalBytes);
		
		// Multiply by 1000 to convert from milliseconds to seconds, and by 8 to convert from bytes to bits
		Float32 bps = (Float32)(intervalBytes * 8) / (Float32)(currentTime - fSession->fLastBitRateSample);
		bps *= 1000;
		fSession->fCurrentBitRate = (UInt32)bps;
		
		// Don't check again for awhile!
		fSession->fLastBitRateSample = currentTime;
	}
	
	for (UInt32 bucketIndex = 0; bucketIndex < fSession->fNumBuckets; bucketIndex++)
	{	
		for (UInt32 bucketMemberIndex = 0; bucketMemberIndex < fSession->fBucketSize; bucketMemberIndex++)
		{	 
			ReflectorOutput* theOutput = fSession->fOutputArray[bucketIndex][bucketMemberIndex];
		
		
			if (theOutput != NULL)
			{	
				SInt32			availBookmarksPosition = -1;	// -1 == invalid position
				OSQueueElem*	packetElem = NULL;				
				UInt32			curBookmark = 0;
								
				Assert( curBookmark < theOutput->fNumBookmarks );		
				
				// see if we've bookmarked a held packet for this Sender in this Output
				while ( curBookmark < theOutput->fNumBookmarks )
				{					
					OSQueueElem* 	bookmarkedElem = theOutput->fBookmarkedPacketsElemsArray[curBookmark];
					
					if ( bookmarkedElem )	// there may be holes in this array
					{							
						if ( bookmarkedElem->IsMember( fPacketQueue ) )	
						{	
							// this packet was previously bookmarked for this specific queue
							// remove if from the bookmark list and use it
							// to jump ahead into the Sender's over all packet queue						
							theOutput->fBookmarkedPacketsElemsArray[curBookmark] = NULL;
							availBookmarksPosition = curBookmark;
							packetElem = bookmarkedElem;
							break;
						}
						
					}
					else
					{
						availBookmarksPosition = curBookmark;
					}
					
					curBookmark++;
						
				}
				
				Assert( availBookmarksPosition != -1 );		
				
				#if REFLECTOR_SESSION_DEBUGGING > 1
				if ( packetElem )	// show 'em what we got johnny
				{	ReflectorPacket* 	thePacket = (ReflectorPacket*)packetElem->GetEnclosingObject();
					printf("Bookmarked packet time: %li, packetSeq %i\n", (long)thePacket->fTimeArrived, DGetPacketSeqNumber( &thePacket->fPacketPtr ) );			
				}
				#endif
				
				// the output did not have a bookmarked packet if it's own
				// so show it the first new packet we have in this sender.
				// ( since TCP flow control may delay the sending of packets, this may not
				// be the same as the first packet in the queue
				if ( packetElem  == NULL )
				{	
					packetElem = fFirstNewPacketInQueue;
						
					#if REFLECTOR_SESSION_DEBUGGING > 1
					if ( packetElem )	// show 'em what we got johnny
					{	
						ReflectorPacket* 	thePacket = (ReflectorPacket*)packetElem->GetEnclosingObject();
						printf("1st NEW packet from Sender sess 0x%lx time: %li, packetSeq %i\n", (long)theOutput, (long)thePacket->fTimeArrived, DGetPacketSeqNumber( &thePacket->fPacketPtr ) );			
					}
					else
						printf("no new packets\n" );
					#endif
				}
				
				OSQueueIter qIter(&fPacketQueue, packetElem);  // starts from beginning if packetElem == NULL, else from packetElem
				
				Bool16			didBookmarkPacket = false;
				UInt32			numPacketsSeen = 0;
				
				while ( !qIter.IsDone() )
				{					
				
					numPacketsSeen++;
					
					packetElem = qIter.GetCurrent();
					
					ReflectorPacket* 	thePacket = (ReflectorPacket*)packetElem->GetEnclosingObject();
					QTSS_Error			err = QTSS_NoErr;
					
					#if REFLECTOR_SESSION_DEBUGGING > 2
					printf("packet time: %li, packetSeq %i\n", (long)thePacket->fTimeArrived, DGetPacketSeqNumber( &thePacket->fPacketPtr ) );			
					#endif
					
					// once we see a packet we cant' send, we need to stop trying
					// during this pass mark remaining as still needed
					if ( !didBookmarkPacket )
					{
						SInt64  packetLateness =  currentTime - thePacket->fTimeArrived - (fSession->fClientSpacing * (SInt64)bucketIndex);
						SInt64	whenToSend; 
					
						whenToSend = this->ThePacketIsToEarlyForTheBucket( thePacket, bucketIndex, currentTime );	// inline function
						
						if ( whenToSend != 0 )
						{	
							#if REFLECTOR_SESSION_DEBUGGING > 2
							printf("packet held: too early packetLateness %li\n" , (long)packetLateness);
							#endif	
							
							// tag it and bookmark it
							thePacket->fNeededByOutput = true;
							
							Assert( availBookmarksPosition != -1 );							
							if ( availBookmarksPosition != -1 )
								theOutput->fBookmarkedPacketsElemsArray[availBookmarksPosition] =  packetElem;
							
							didBookmarkPacket = true;
							
							// adjust our next run time to match the
							// packet that needs to go soonest
							if ( fNextTimeToRun == -1 || fNextTimeToRun > whenToSend )
								fNextTimeToRun = whenToSend;
						}				
						else
						{
							// packetLateness measures how late this packet it after being corrected for the bucket delay
							
							#if REFLECTOR_SESSION_DEBUGGING > 2
							printf("packetLateness %li, seq# %li\n", (long)packetLateness, (long) DGetPacketSeqNumber( &thePacket->fPacketPtr ) );			
							#endif
						
							err = theOutput->WritePacket(&thePacket->fPacketPtr, fStream->fStreamInfo.fTrackID, fWriteFlag, packetLateness );
							
							if ( err == QTSS_WouldBlock )
							{	
								#if REFLECTOR_SESSION_DEBUGGING 
								printf("EAGAIN bookmark: %li, packetSeq %i, cur queue len: %li\n", (long)packetLateness, DGetPacketSeqNumber( &thePacket->fPacketPtr ),  (long)fPacketQueue.GetLength()  );			
								#endif
								// tag it and bookmark it
								thePacket->fNeededByOutput = true;
								
								Assert( availBookmarksPosition != -1 );
								if ( availBookmarksPosition != -1 )
									theOutput->fBookmarkedPacketsElemsArray[availBookmarksPosition] =  packetElem;

								didBookmarkPacket = true;
								
								// call us again in # ms to retry on an EAGAIN
								if ( fNextTimeToRun > 10 )
									fNextTimeToRun = 10;
							}
							
						}
					}
					else
					{	
						if ( thePacket->fNeededByOutput == true )	// optimization: if the packet is already marked, another Output has been through this already
							break;
						thePacket->fNeededByOutput = true;
					}
					
					qIter.Next();
				} 
				
				#if REFLECTOR_SESSION_DEBUGGING >2
				if ( printQueueLenOnExit && numPacketsSeen > 0)
				{
					printf("packets seen by output(%li): %li\n", (long)(bucketIndex+1)*(bucketMemberIndex+1), numPacketsSeen );			
					
				}
				#endif
				
			}

		}
	}
	
	// reset our first new packet bookmark
	fFirstNewPacketInQueue = NULL;


	if ( isTimeToCleanUpQueue )
	{
		SInt32 	packetsRequeued = 0;
		SInt32 	packetsHeld = 0;
		
		// iterate one more through the senders queue to clear out
		// the unneeded packets
		OSQueueIter removeIter(&fPacketQueue);
		while ( !removeIter.IsDone() )
		{
			OSQueueElem* elem = removeIter.GetCurrent();
			Assert( elem );

			//at this point, move onto the next queue element, because we may be altering
			//the queue itself in the code below
			removeIter.Next();

			ReflectorPacket* thePacket = (ReflectorPacket*)elem->GetEnclosingObject();		
			Assert( thePacket );
			
			if ( thePacket->fNeededByOutput == false )
			{	
				thePacket->fNeededByOutput = true;
				fPacketQueue.Remove( elem );
				inFreeQueue->EnQueue( elem );
				
				packetsRequeued++;
				
				
			}
			else	// reset for next call to ReflectPackets
			{
				thePacket->fNeededByOutput = false;
				packetsHeld++;
			}
			
			
		}
		
		#if REFLECTOR_SESSION_DEBUGGING
		if ( packetsRequeued )
			printf(" %li packet released, packetsHeld %li\n", packetsRequeued, packetsHeld );	
		#endif
	}
	
	//Don't forget that the caller also wants to know when we next want to run
	if ( fNextTimeToRun == -1 )	// reset from never
		fNextTimeToRun = 0;
		
	if (*ioWakeupTime == 0)
		*ioWakeupTime = fNextTimeToRun;
	else if ((fNextTimeToRun > 0) && (*ioWakeupTime > fNextTimeToRun))
		*ioWakeupTime = fNextTimeToRun;
	// exit with fNextTimeToRun in real time, not relative time.
	fNextTimeToRun += currentTime;
	
}


#pragma mark __REFLECTOR_SOCKET__

UDPSocketPair* ReflectorSocketPool::ConstructUDPSocketPair()
{
	return new ('usce') UDPSocketPair(new ('rsck') ReflectorSocket(),
										new ('rsck') ReflectorSocket());
}

void ReflectorSocketPool::DestructUDPSocketPair(UDPSocketPair *inPair)
{
	//The socket's run function may be executing RIGHT NOW! So we can't
	//just delete the thing, we need to send the sockets kill events.
	((ReflectorSocket*)inPair->GetSocketA())->Signal(Task::kKillEvent);
	((ReflectorSocket*)inPair->GetSocketB())->Signal(Task::kKillEvent);
	delete inPair;
}

ReflectorSocket::ReflectorSocket()
: IdleTask(), UDPSocket(this, Socket::kNonBlockingSocketType), fMutex('rsss'), fSleepTime(0)
{
	//construct all the preallocated packets
	for (UInt32 numPackets = 0; numPackets < kNumPreallocatedPackets; numPackets++)
	{
		//If the local port # of this socket is odd, then all the packets
		//used for this socket are rtcp packets.
		ReflectorPacket* packet = new ('rpka') ReflectorPacket();
		fFreeQueue.EnQueue(&packet->fQueueElem);//put this packet onto the free queue
	}
}

ReflectorSocket::~ReflectorSocket()
{
	while (fFreeQueue.GetLength() > 0)
	{
		ReflectorPacket* packet = (ReflectorPacket*)fFreeQueue.DeQueue()->GetEnclosingObject();
		delete packet;
	}
}

void	ReflectorSocket::AddSender(ReflectorSender* inSender)
{
	OSMutexLocker locker(&fMutex);
	fSenderQueue.EnQueue(&inSender->fSocketQueueElem);
}

void	ReflectorSocket::RemoveSender(ReflectorSender* inSender)
{
	OSMutexLocker locker(&fMutex);
	fSenderQueue.Remove(&inSender->fSocketQueueElem);
}

SInt64 ReflectorSocket::Run()
{
	//We want to make sure we can't get idle events WHILE we are inside
	//this function. That will cause us to run the queues unnecessarily
	//and just get all confused.
	this->CancelTimeout();
	
	Task::EventFlags theEvents = this->GetEvents();
	//if we have been told to delete ourselves, do so.
	if (theEvents & Task::kKillEvent)
		return -1;

	OSMutexLocker locker(&fMutex);
	SInt64 theMilliseconds = OS::Milliseconds();
	
	//Only check for data on the socket if we've actually been notified to that effect
	if (theEvents & Task::kReadEvent)
		this->GetIncomingData(theMilliseconds);

#if DEBUG
	//make sure that we haven't gotten here prematurely! This wouldn't mess
	//anything up, but it would waste CPU.
	if (theEvents & Task::kIdleEvent)
	{
		SInt32 temp = (SInt32)(fSleepTime - theMilliseconds);
		char tempBuf[20];
		::sprintf(tempBuf,"%ld",temp);
		WarnV(fSleepTime <= theMilliseconds, tempBuf);
	}
#endif
	
	fSleepTime = 0;
	//Now that we've gotten all available packets, have the streams reflect
	for (OSQueueIter iter2(&fSenderQueue); !iter2.IsDone(); iter2.Next())
	{
		ReflectorSender* theSender2 = (ReflectorSender*)iter2.GetCurrent()->GetEnclosingObject();
		if (theSender2->ShouldReflectNow(theMilliseconds, &fSleepTime))
			theSender2->ReflectPackets(&fSleepTime, &fFreeQueue);
#if DEBUG
		theSender2->fHasDroppedDupPacket = false;
#endif
	}
	
#if DEBUG
	theMilliseconds = OS::Milliseconds();
#endif
	
	//For smoothing purposes, the streams can mark when they want to wakeup.
	if (fSleepTime > 0)
		this->SetIdleTimer(fSleepTime);
#if DEBUG
	//The debugging check above expects real time.
	fSleepTime += theMilliseconds;
#endif
		
	return 0;
}

void ReflectorSocket::GetIncomingData(const SInt64& inMilliseconds)
{
	UInt32 theMulticastAddr = 0;
	UInt32 theRemoteAddr = 0;
	UInt16 theRemotePort = 0;
#if REFLECTOR_SESSION_DEBUGGING > 0	
	UInt32	numPackets = 0;
#endif
	
	//get all the outstanding packets for this socket
	while (true)
	{
		//get a packet off the free queue.
		ReflectorPacket* thePacket = this->GetPacket();

		thePacket->fPacketPtr.Len = 0;
		(void)this->RecvPacket(&theMulticastAddr, &theRemoteAddr, &theRemotePort, thePacket->fPacketPtr.Ptr,
							ReflectorPacket::kMaxReflectorPacketSize, &thePacket->fPacketPtr.Len);
		if (thePacket->fPacketPtr.Len == 0)
		{
			//put the packet back on the free queue, because we didn't actually
			//get any data here.
			fFreeQueue.EnQueue(&thePacket->fQueueElem);
			this->RequestEvent(EV_RE);
			break;//no more packets on this socket!
		}
		
#if REFLECTOR_SESSION_DEBUGGING > 0	
		numPackets++;
#endif		
		if (GetLocalPort() & 1)
		{
			//if this is a new RTCP packet, check to see if it is a sender report.
			//We should only reflect sender reports. Because RTCP packets can't have both
			//an SR & an RR, and because the SR & the RR must be the first packet in a
			//compound RTCP packet, all we have to do to determine this is look at the
			//packet type of the first packet in the compound packet.
			RTCPPacket theRTCPPacket;
			if ((!theRTCPPacket.ParsePacket((UInt8*)thePacket->fPacketPtr.Ptr, thePacket->fPacketPtr.Len)) ||
				(theRTCPPacket.GetPacketType() != RTCPSRPacket::kSRPacketType))
			{
				//pretend as if we never got this packet
				fFreeQueue.EnQueue(&thePacket->fQueueElem);
				continue;
			}
		}
	
		//We're going to have to reflect this packet, so put it on the proper queue
		Bool16 hasStream = false;
		
		//find the right stream for this packet
		for (OSQueueIter iter(&fSenderQueue); (!iter.IsDone() && !hasStream); iter.Next())
		{
			ReflectorSender* theSender = (ReflectorSender*)iter.GetCurrent()->GetEnclosingObject();
			if ((theSender->fStream->fStreamInfo.fDestIPAddr == theMulticastAddr) || (theMulticastAddr == 0))
			{
				//Here we're trying to route packets to the right reflector stream
				//based on destination (multicast) addr. If the OS doesn't support
				//this feature, the above function will return 0 for the addr.
#if OS_SUPPORTS_UDP_DEST_ADDR
				Assert(theMulticastAddr != 0);
#endif

				//Before putting this on the sender's packet queue, make sure this isn't
				//a duplicate of a previous packet. Only do this for RTP packets, using
				//the RTP sequence number
				if (!(this->GetLocalPort() & 1))
				{
					UInt16* theSeqNumberP = (UInt16*)thePacket->fPacketPtr.Ptr;
					
					//AddSequenceNumber returns true if we've seen this sequence number before
					if (theSender->fStream->fSequenceNumberMap.AddSequenceNumber(ntohs(theSeqNumberP[1])))
					{
						//WarnV(0, "Found duplicate seq #\n");// egads... is this causing problems?
#if DEBUG
						theSender->fHasDroppedDupPacket = true;
#endif
						break;
					}
					
					// Because this is an RTP packet, and it's not a duplicate, lets update the
					// ReflectorSession's current byte count. Make sure to atomic add this because
					// multiple sockets can be adding to this variable simultaneously
					(void)atomic_add(&theSender->fSession->fBytesSentInThisInterval, thePacket->fPacketPtr.Len);
				}
				
				// Check to see if we need to set the remote RTCP address
				// for this stream. This will be necessary if the source is unicast.
				if ((theRemoteAddr != 0) && (theSender->fStream->fDestRTCPAddr == 0))
				{
					// If the source is multicast, this shouldn't be necessary
					Assert(!SocketUtils::IsMulticastIPAddr(theSender->fStream->fStreamInfo.fDestIPAddr));
					Assert(theRemotePort != 0);
					theSender->fStream->fDestRTCPAddr = theRemoteAddr;
					theSender->fStream->fDestRTCPPort = theRemotePort;
					
					// RTCPs are always on odd ports, so check to see if this port is an
					// RTP port, and if so, just add 1.
					if (!(theRemotePort & 1))
						theSender->fStream->fDestRTCPPort++;				
				}

				thePacket->fBucketsSeenThisPacket = 0;
				thePacket->fTimeArrived = inMilliseconds;
				theSender->fPacketQueue.EnQueue(&thePacket->fQueueElem);			
				if ( theSender->fFirstNewPacketInQueue == NULL )
					theSender->fFirstNewPacketInQueue = &thePacket->fQueueElem;					
				theSender->fHasNewPackets = true;
				hasStream = true;
			}
		}
		//if none of the Reflector streams wanted this packet, make sure not to leak it!
		if (!hasStream)
			fFreeQueue.EnQueue(&thePacket->fQueueElem);
	}
	
#if REFLECTOR_SESSION_DEBUGGING > 0	
	if ( numPackets )
		printf( "numPackets read by reflector %i\n", numPackets );
#endif
}


ReflectorPacket* ReflectorSocket::GetPacket()
{
	if (fFreeQueue.GetLength() == 0)
		//if the port number of this socket is odd, this packet is an RTCP packet.
		return new ('rpka') ReflectorPacket();
	else
		return (ReflectorPacket*)fFreeQueue.DeQueue()->GetEnclosingObject();
}
