/*
 * Copyright (c) 1999 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * Copyright (c) 1999 Apple Computer, Inc.  All Rights Reserved.
 * The contents of this file constitute Original Code as defined in and are 
 * subject to the Apple Public Source License Version 1.1 (the "License").  
 * You may not use this file except in compliance with the License.  Please 
 * obtain a copy of the License at http://www.apple.com/publicsource and 
 * read it before using this file.
 * 
 * This Original Code and all software distributed under the License are 
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, EITHER 
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES, 
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY, FITNESS 
 * FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT.  Please see the License for 
 * the specific language governing rights and limitations under the 
 * License.
 * 
 * 
 * @APPLE_LICENSE_HEADER_END@
 */
/*
	File:		ReflectorSession.h

	Contains:	This object supports reflecting an RTP multicast stream to N
				RTPStreams. It spaces out the packet send times in order to
				maximize the randomness of the sending pattern and smooth
				the stream.
					
	$Log: ReflectorSession.h,v $
	Revision 1.10  1999/12/21 01:43:26  lecroy
	*** empty log message ***
	
	Revision 1.9  1999/10/28 20:59:20  serenyi
	Got rid of dup packet WarnV, modified wakeup time assert
	
	Revision 1.8  1999/10/28 01:45:25  serenyi
	Fix 2404454 ReflectorSession & module were still assuming trackIDs were always 1-based
	
	Revision 1.7  1999/10/26 23:57:03  rtucker
	enhanced reflector performance with large packet queues
	
	Revision 1.6  1999/10/21 18:30:06  rtucker
	fixed "double send" bug in relfector
	
	Revision 1.5  1999/10/19 22:17:28  serenyi
	ReflectorSession now keeps track of its bitrate, each client puts that bitrate into the RTPSession
	
	Revision 1.4  1999/10/16 00:57:28  rtucker
	fix reflector bugs from thinning
	
	Revision 1.3  1999/10/14 01:17:29  rtucker
	updates for thinning
	
	Revision 1.2  1999/10/07 09:01:50  serenyi
	Changed bool to Bool16, changed QTSS_Dictionary to QTSS_Object
	
	Revision 1.1.1.1  1999/09/22 23:01:08  lecroy
	importing
	
	Revision 1.2  1999/09/15 17:01:32  serenyi
	Made relay, reflector work
	
	Revision 1.1.1.1  1999/09/13 16:16:34  rtucker
	initial checkin from rrv14
	
	
	Revision 1.1  1999/02/19 23:10:05  ds
	Created
	

*/

#ifndef _REFLECTOR_SESSION_H_
#define _REFLECTOR_SESSION_H_

#include "QTSS.h"

#include "IdleTask.h"
#include "SourceInfo.h"

#include "UDPSocket.h"
#include "UDPSocketPool.h"
#include "EventContext.h"
#include "SequenceNumberMap.h"

#include "OSMutex.h"
#include "OSQueue.h"
#include "StrPtrLen.h"
#include "ResizeableStringFormatter.h"

#include "RTCPSRPacket.h"
#include "ReflectorOutput.h"

//This will add some printfs that are useful for checking the thinning
#define REFLECTOR_THINNING_DEBUGGING 0 


class ReflectorStream;
class ReflectorBuffer;
class ReflectorPacket;
class ReflectorSender;

class ReflectorSession
{
	public:
	
		// Call this once at startup
		static void Register();
	
		//This object gets scheduled and sends out packets
		
		ReflectorSession(SourceInfo* inInfo, UInt32 inBucketDelayInMsec, UInt32 inBucketSize);
		~ReflectorSession();
		
		//
		// MODIFIERS
		
		// Socket setup & teardown
		QTSS_Error		BindReflectorSockets(QTSS_RTSPRequestObject inRequest);
		
		//RTP sessions can attach and unattach to this reflector session as they
		//wish by calling these methods
		void	AddOutput(ReflectorOutput* inOutput);
		void 	RemoveOutput(ReflectorOutput* inOutput);
		
		//
		// LOCAL SDP DATA
		// The SourceInfo this ReflectorSession is based on may or may not be
		// able to generate LocalSDP.
		void	UpdateLocalSDP(SourceInfo* inInfo);

		//
		// ACCESSORS
		
		OSMutex*		GetMutex() 	{ return &fMutex; }
		OSRef*			GetRef()	{ return &fRef; }
		UInt32			GetNumStreams() { return fNumStreams; }
		UInt32			GetNumOutputs() { return fNumElements; }
		StrPtrLen*		GetSourceInfoHTML() { return &fSourceInfoHTML; }
		
		// A ReflectorSession keeps track of the aggregate bit rate each
		// stream is reflecting (RTP only). Initially, this will return 0
		// until enough time passes to compute an accurate average.
		UInt32			GetCurrentBitsPerSec() { return fCurrentBitRate; }
		
		StrPtrLen*			GetPayloadName(UInt32 trackID);
		QTSS_RTPPayloadType	GetPayloadType(UInt32 trackID);
		
		// Returns the Local SDP, if available
		StrPtrLen*	GetLocalSDP() { return &fLocalSDP; }
		
		// Returns true if this session is Equal to the SourceInfo
		Bool16 Equal(SourceInfo* inInfo);
	
		//Reflector quality levels:
		enum
		{
			kAudioOnlyQuality = 1,		//UInt32
			kNormalQuality = 0,			//UInt32
			kNumQualityLevels = 2		//UInt32
		};

	private:
	
		// Helper methods
		void 		AllocateSessionArray(UInt32 inNumBuckets);
		void		CloseSourceSockets();


		
		//Data structures can't be modified while we are sending packets.
		OSMutex		fMutex;
	
		ReflectorStream** 	fStreamArray;
		UInt32				fNumStreams;

		// For storage in the session map		
		OSRef  	fRef;
		// Each ReflectorSession is identified in the map by the SourceInfo unique ID
		char	fSourceIDBuf[SourceInfo::kSourceIDSize];
		StrPtrLen fSourceID;

		//RTPSessions are kept in a 2-dimensional array
		enum
		{
			kMinNumBuckets = 64,	//UInt32
			kMaxHTMLSize = 128,
			kBitRateAvgIntervalInMilSecs = 30000 // time between bitrate averages
		};
		typedef ReflectorOutput** Bucket;
		Bucket*		fOutputArray;

		UInt32		fBucketSize;
		UInt32		fNumBuckets;
		UInt32		fFirstEmptyBucket;//first bucket that's completely empty
	
		//time (in msec) between each client sending out their packets
		SInt64		fClientSpacing;

		UInt32		fNumElements;
		StrPtrLen	fLocalSDP;
		
		char		fHTMLBuf[kMaxHTMLSize];
		StrPtrLen	fSourceInfoHTML;
		ResizeableStringFormatter fFormatter;
		
		// Used for calculating average bit rate
		UInt32				fCurrentBitRate;
		SInt64				fLastBitRateSample;
		unsigned int		fBytesSentInThisInterval;// unsigned long because we need to atomic_add it

		friend class ReflectorStream;
		friend class ReflectorSender;
		friend class ReflectorSocket;
};

class ReflectorPacket
{
	private:
	
		ReflectorPacket() : fQueueElem(this), fPacketPtr(fPacketData, 0), fNeededByOutput(true) {}
		~ReflectorPacket() {}

		enum
		{
			kMaxReflectorPacketSize = 2048	//UInt32
		};

		UInt32 		fBucketsSeenThisPacket;
		SInt64 		fTimeArrived;
		OSQueueElem	fQueueElem;
		char		fPacketData[kMaxReflectorPacketSize];
		StrPtrLen	fPacketPtr;
		
		Bool16		fNeededByOutput; // is this packet still needed for output?
				
		friend class ReflectorSender;
		friend class ReflectorSocket;
};


//Custom UDP socket classes for doing reflector packet retrieval, socket management
class ReflectorSocket : public IdleTask, public UDPSocket
{
	public:

		ReflectorSocket();
		virtual ~ReflectorSocket();
		
		void	AddSender(ReflectorSender* inSender);
		void	RemoveSender(ReflectorSender* inStreamElem);
		Bool16	HasSender() { return (fSenderQueue.GetLength() > 0); }
	
	private:
		
		virtual SInt64		Run();
		ReflectorPacket* 	GetPacket();
		void 				GetIncomingData(const SInt64& inMilliseconds);
		
		//Number of packets to allocate when the socket is first created
		enum
		{
			kNumPreallocatedPackets = 20	//UInt32
		};

		//Queue of available ReflectorPackets
		OSQueue	fFreeQueue;
		//Queue of attached senders
		OSQueue	fSenderQueue;
		//Can't have new streams attaching to this socket while we
		//are processing data. We could get rid of this mutex if we used an atomic queue
		OSMutex	fMutex;
		
		SInt64	fSleepTime;
};


class ReflectorSocketPool : public UDPSocketPool
{
	public:
	
		ReflectorSocketPool() {}
		virtual ~ReflectorSocketPool() {}
		
		virtual UDPSocketPair*	ConstructUDPSocketPair();
		virtual void			DestructUDPSocketPair(UDPSocketPair *inPair);
};

class ReflectorSender
{
	ReflectorSender(ReflectorStream* inStream, ReflectorSession* inSession, UInt32 inWriteFlag);
	~ReflectorSender();
	
	//Used for adjusting sequence numbers in light of thinning
	UInt16		GetPacketSeqNumber(const StrPtrLen& inPacket);
	void		SetPacketSeqNumber(const StrPtrLen& inPacket, UInt16 inSeqNumber);
	Bool16 		PacketShouldBeThinned(QTSS_RTPStreamObject inStream, const StrPtrLen& inPacket);

	//We want to make sure that ReflectPackets only gets invoked when there
	//is actually work to do, because it is an expensive function
	Bool16		ShouldReflectNow(const SInt64& inCurrentTime, SInt64* ioWakeupTime);
	
	//This function gets data from the multicast source and reflects.
	//Returns the time at which it next needs to be invoked
	void		ReflectPackets(SInt64* ioWakeupTime, OSQueue* inFreeQueue);

	ReflectorStream* 	fStream;
	ReflectorSession* 	fSession;
	UInt32				fWriteFlag;
	
	OSQueueElem	fSocketQueueElem;
	
	OSQueue			fPacketQueue;
	OSQueueElem*	fFirstNewPacketInQueue;

	//these serve as an optimization, keeping track of when this
	//sender needs to run so it doesn't run unnecessarily
	Bool16		fHasNewPackets;
#if DEBUG
	Bool16		fHasDroppedDupPacket;
#endif

	SInt64		fNextTimeToRun;
			
	//how often to send RRs to the source
	enum
	{
		kRRInterval = 5000,		//SInt64 (every 5 seconds)
		kCleanUpInterval = 2000,		//SInt64 (every # MSec)
		kSmallPacketCutoff = 200//UInt32
	};

	SInt64		fLastRRTime;
	SInt64		fLastQueueCleanupTime;
	
	
	inline SInt64		ThePacketIsToEarlyForTheBucket( ReflectorPacket* thePacket, UInt32 bucketIndex, SInt64 currentTimeInMSec );
	
	friend class ReflectorSocket;
	friend class ReflectorStream;
};


class ReflectorStream
{
	public:
	
		ReflectorStream(ReflectorSession* inSession, SourceInfo::StreamInfo* inInfo);
		~ReflectorStream()
		{ if (fStreamInfo.fPayloadName.Ptr != NULL) delete [] fStreamInfo.fPayloadName.Ptr; }
		
		//Call this to initialize the reflector sockets. Returns an RTSPHandler
		//in the event of an error
		QTSS_Error	BindSockets(QTSS_RTSPRequestObject inRequest);
		void			CloseSockets();
		
		//Sends an RTCP receiver report to the multicast source
		void	SendReceiverReport();
		
		StrPtrLen*				GetPayloadName() 	{ return &fStreamInfo.fPayloadName; }
		QTSS_RTPPayloadType		GetPayloadType()	{ return fStreamInfo.fPayloadType; }
		
	private:
	
		UDPSocketPair*		fSockets;
		ReflectorSession*	fSession;
		ReflectorSender		fRTPSender;
		ReflectorSender		fRTCPSender;
		SequenceNumberMap	fSequenceNumberMap; //for removing duplicate packets
		
		// All the necessary info about this stream
		SourceInfo::StreamInfo	fStreamInfo;
		
		//for issuing RTCP RRs to source
		enum
		{
			kReceiverReportSize = 16,	//UInt32
			kAppSize = 36				//UInt32
		};
	
		char		fReceiverReportBuffer[kReceiverReportSize + kAppSize +
										RTCPSRPacket::kMaxCNameLen];
		UInt32*		fEyeLocation;//place in the buffer to write the eye information
		UInt32		fReceiverReportSize;
		
		// This is the destination address & port for RTCP
		// receiver reports.
		UInt32		fDestRTCPAddr;
		UInt16		fDestRTCPPort;
		
		static ReflectorSocketPool	sSocketPool;
		
		friend class ReflectorSession;
		friend class ReflectorSocket;
		friend class ReflectorSender;
};

// return 0 if this packet is within the time window to send
// else return # MSecs to wait before sending this packet
SInt64	ReflectorSender::ThePacketIsToEarlyForTheBucket( ReflectorPacket* thePacket, UInt32 bucketIndex, SInt64 currentTimeInMSec )
{
	// each successive bucket is delayed by fSession->fClientSpacing
	// this smooths the bandwidth hump caused by the output of the rebroadcast of new packets.
	
	SInt64	whenToSend;
	
	whenToSend = thePacket->fTimeArrived + ( (SInt64)bucketIndex * fSession->fClientSpacing )  - currentTimeInMSec;
	
	if ( whenToSend < 0 )
	{	
		//printf("Packet fTimeArrived: %li, bucketIndex %li, fClientSpacing %li OS::Milliseconds() %li\n", (long)thePacket->fTimeArrived, bucketIndex, (long)fSession->fClientSpacing, (long)OS::Milliseconds());
		return 0;
	
	}

	return whenToSend;

}

#endif //_REFLECTOR_SESSION_H_

