/*
 * Copyright (c) 1999 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * Copyright (c) 1999 Apple Computer, Inc.  All Rights Reserved.
 * The contents of this file constitute Original Code as defined in and are 
 * subject to the Apple Public Source License Version 1.1 (the "License").  
 * You may not use this file except in compliance with the License.  Please 
 * obtain a copy of the License at http://www.apple.com/publicsource and 
 * read it before using this file.
 * 
 * This Original Code and all software distributed under the License are 
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, EITHER 
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES, 
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY, FITNESS 
 * FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT.  Please see the License for 
 * the specific language governing rights and limitations under the 
 * License.
 * 
 * 
 * @APPLE_LICENSE_HEADER_END@
 */
/*
	File:		OSMemory_Server.cpp

	Contains:	Implementation of OSMemory stuff, including all new & delete
				operators.
					
	$Log: OSMemory.cpp,v $
	Revision 1.5  1999/12/21 01:43:40  lecroy
	*** empty log message ***
	
	Revision 1.4  1999/10/27 17:54:07  serenyi
	Integrated Matthias's MacOS / GUSI changes
	
	Revision 1.3  1999/10/07 09:02:56  serenyi
	Changed bool to Bool16, changed QTSS_Dictionary to QTSS_Object
	
	Revision 1.2  1999/09/30 00:27:05  serenyi
	Always call ValidateMemoryQueue when memory debugging
	
	Revision 1.1.1.1  1999/09/22 23:01:25  lecroy
	importing
	
	Revision 1.2  1999/09/17 22:57:57  serenyi
	Got rid of ValidateMemoryQueue checks, as these are too expensive
	
	Revision 1.1.1.1  1999/09/13 16:16:40  rtucker
	initial checkin from rrv14
	
	

	Created: Tue, Mar 2, 1999 @ 4:23 PM
*/

#include <string.h>
#include "OSMemory.h"

#if MEMORY_DEBUGGING

OSQueue OSMemory::sMemoryQueue;
OSQueue OSMemory::sTagQueue;
UInt32  OSMemory::sAllocatedBytes = 0;
OSMutex  OSMemory::sMutex('Memy');

#endif

static SInt32 	sMemoryErr = 0;

void* operator new (size_t s)
{
	return OSMemory::New (s, '0000', false);
}

void* operator new[](size_t s)
{
	return OSMemory::New (s, '0000', false);
}

void* operator new(size_t s, UInt32 tag)
{
	return OSMemory::New (s, tag, true);
}

void* operator new[](size_t s, UInt32 tag)
{
	return OSMemory::New (s, tag, false);
}

void operator delete(void* mem)
{
	OSMemory::Delete(mem);
}


void operator delete[](void* mem)
{
	operator delete(mem);
}

void OSMemory::SetMemoryError(SInt32 inErr)
{
	sMemoryErr = inErr;
}

#if MEMORY_DEBUGGING
void* OSMemory::New(size_t s, UInt32 tag, Bool16 sizeCheck)
#else
void* OSMemory::New(size_t s, UInt32 /*tag*/, Bool16 /*sizeCheck*/)
#endif
{
#if MEMORY_DEBUGGING
	return OSMemory::DebugNew(s, tag, sizeCheck);
#else
	void *m = malloc(s);
	if (m == NULL)
		exit(sMemoryErr);
	return m;
#endif
}

void OSMemory::Delete(void* inMemory)
{
#if MEMORY_DEBUGGING
	OSMemory::DebugDelete(inMemory);
#else
	if (inMemory == NULL)
		return;
	free(inMemory);
#endif
}

#if MEMORY_DEBUGGING
void* OSMemory::DebugNew(size_t s, UInt32 tag, Bool16 sizeCheck)
{
	//also allocate enough space for a Q elem and a long to store the length of this
	//allocation block
	OSMutexLocker locker(&sMutex);
	ValidateMemoryQueue();
	UInt32 actualSize = s + sizeof(MemoryDebugging) + (2 * sizeof(tag));
	char *m = (char *)malloc(actualSize);
	if (m == NULL)
		exit(sMemoryErr);

	//mark the beginning and the end with the tag
	memset(m, 0xfe, actualSize);//mark the block with an easily identifiable pattern
	memcpy(m, &tag, sizeof(tag));
	memcpy((m + actualSize) - sizeof(tag), &tag, sizeof(tag));

	//put this chunk on the global chunk queue
	MemoryDebugging* header = (MemoryDebugging*)(m + sizeof(tag));
	memset(header, 0, sizeof(MemoryDebugging));
	header->size = s;
	header->tag = tag;
	header->elem.SetEnclosingObject(header);
	sMemoryQueue.EnQueue(&header->elem);
	sAllocatedBytes += s;
	
	//also update the tag queue
	for (OSQueueIter iter(&sTagQueue); !iter.IsDone(); iter.Next())
	{
		TagElem* elem = (TagElem*)iter.GetCurrent()->GetEnclosingObject();
		if (*elem->tagPtr == tag)
		{
			//verify that the size of this allocation is the same as all others
			//(if requested... some tags are of variable size)
			if (sizeCheck)
				Assert(s == elem->tagSize);
			elem->totMemory += s;
			elem->numObjects++;
			return header + 1;
		}
	}
	//if we've gotten here, this tag doesn't exist, so let's add it.
	TagElem* elem = (TagElem*)malloc(sizeof(TagElem));
	if (elem == NULL)
		exit(sMemoryErr);
	memset(elem, 0, sizeof(TagElem));
	elem->elem.SetEnclosingObject(elem);
	memcpy(elem->tag, &tag, sizeof(tag));
	elem->tagPtr = (UInt32*)&elem->tag;
	elem->tagSize = s;
	elem->totMemory = s;
	elem->numObjects = 1;
	sTagQueue.EnQueue(&elem->elem);
	return m + sizeof(tag) + sizeof(MemoryDebugging);
}

void OSMemory::DebugDelete(void *mem)
{
	OSMutexLocker locker(&sMutex);
	ValidateMemoryQueue();
	char* memPtr = (char*)mem;
	MemoryDebugging* memInfo = (MemoryDebugging*)mem;
	memInfo--;//get a pointer to the MemoryDebugging structure
	Assert(memInfo->elem.IsMemberOfAnyQueue());//must be on the memory Queue
	//double check it's on the memory queue
	Bool16 found  = false;
	for (OSQueueIter iter(&sMemoryQueue); !iter.IsDone(); iter.Next())
	{
		MemoryDebugging* check = (MemoryDebugging*)iter.GetCurrent()->GetEnclosingObject();
		if (check == memInfo)
		{
			found = true;
			break;
		}
	}
	Assert(found == true);
	sMemoryQueue.Remove(&memInfo->elem);
	Assert(!memInfo->elem.IsMemberOfAnyQueue());
	sAllocatedBytes -= memInfo->size;
	
	//verify that the tags placed at the very beginning and very end of the
	//block still exist
	memPtr += memInfo->size;
	UInt32* tagPtr = (UInt32*)memPtr;
	Assert(*tagPtr == memInfo->tag);
	memPtr -= sizeof(MemoryDebugging) + sizeof(UInt32) + memInfo->size;
	tagPtr = (UInt32*)memPtr;
	Assert(*tagPtr == memInfo->tag);
	
	//also update the tag queue
	for (OSQueueIter iter2(&sTagQueue); !iter2.IsDone(); iter2.Next())
	{
		TagElem* elem = (TagElem*)iter2.GetCurrent()->GetEnclosingObject();
		if (*elem->tagPtr == memInfo->tag)
		{
			Assert(elem->numObjects > 0);
			elem->numObjects--;
			elem->totMemory -= memInfo->size;
			memset(mem, 0xfd,memInfo->size);
			free(memPtr);
			return;
		}
	}
	Assert(0);//we should always find this tag on the tag queue.
}

void OSMemory::ValidateMemoryQueue()
{
	OSMutexLocker locker(&sMutex);
	for(OSQueueIter iter(&sMemoryQueue); !iter.IsDone(); iter.Next())
	{
		MemoryDebugging* elem = (MemoryDebugging*)iter.GetCurrent()->GetEnclosingObject();
		char* rawmem = (char*)elem;
		rawmem -= sizeof(UInt32);
		UInt32* tagPtr = (UInt32*)rawmem;
		rawmem += sizeof(UInt32) + sizeof(MemoryDebugging) + elem->size;
		tagPtr = (UInt32*)rawmem;
		Assert(*tagPtr == elem->tag);
	}
	for(OSQueueIter iter2(&sMemoryQueue); !iter2.IsDone(); iter2.Next())
	{
		MemoryDebugging* elem2 = (MemoryDebugging*)iter2.GetCurrent()->GetEnclosingObject();
		char* rawmem2 = (char*)elem2;
		rawmem2 -= sizeof(UInt32);
		UInt32* tagPtr2 = (UInt32*)rawmem2;
		Assert(*tagPtr2 == elem2->tag);
	}
}


Bool16 OSMemory::MemoryDebuggingTest()
{
	static char* s20 = "this is 20 characte";
	static char* s30 = "this is 30 characters long, o";
	static char* s40 = "this is 40 characters long, okey dokeys";
	
	void* victim = DebugNew(20, 'tsta', true);
	strcpy((char*)victim, s20);
	MemoryDebugging* victimInfo = (MemoryDebugging*)victim;
	ValidateMemoryQueue();
	victimInfo--;
	if (victimInfo->tag != 'tsta')
		return false;
	if (victimInfo->size != 20)
		return false;
		
	void* victim2 = DebugNew(30, 'tstb', true);
	strcpy((char*)victim2, s30);
	ValidateMemoryQueue();
	void* victim3 = DebugNew(20, 'tsta', true);
	strcpy((char*)victim3, s20);
	ValidateMemoryQueue();
	void* victim4 = DebugNew(40, 'tstc', true);
	strcpy((char*)victim4, s40);
	ValidateMemoryQueue();
	void* victim5 = DebugNew(30, 'tstb', true);
	strcpy((char*)victim5, s30);
	ValidateMemoryQueue();
	
	if (sTagQueue.GetLength() != 3)
		return false;
	for (OSQueueIter iter(&sTagQueue); !iter.IsDone(); iter.Next())
	{
		TagElem* elem = (TagElem*)iter.GetCurrent()->GetEnclosingObject();
		if (*elem->tagPtr == 'tstb')
		{
			if (elem->tagSize != 30)
				return false;
			if (elem->numObjects != 2)
				return false;
		}
		else if (*elem->tagPtr == 'tsta')
		{
			if (elem->tagSize != 20)
				return false;
			if (elem->numObjects != 2)
				return false;
		}
		else if (*elem->tagPtr == 'tstc')
		{
			if (elem->tagSize != 40)
				return false;
			if (elem->numObjects != 1)
				return false;
		}
		else
			return false;
	}
	
	DebugDelete(victim3);
	ValidateMemoryQueue();
	DebugDelete(victim4);
	ValidateMemoryQueue();

	if (sTagQueue.GetLength() != 3)
		return false;
	for (OSQueueIter iter2(&sTagQueue); !iter2.IsDone(); iter2.Next())
	{
		TagElem* elem = (TagElem*)iter2.GetCurrent()->GetEnclosingObject();
		if (*elem->tagPtr == 'tstb')
		{
			if (elem->tagSize != 30)
				return false;
			if (elem->numObjects != 2)
				return false;
		}
		else if (*elem->tagPtr == 'tsta')
		{
			if (elem->tagSize != 20)
				return false;
			if (elem->numObjects != 1)
				return false;
		}
		else if (*elem->tagPtr == 'tstc')
		{
			if (elem->tagSize != 40)
				return false;
			if (elem->numObjects != 0)
				return false;
		}
		else
			return false;
	}
	
	if (sMemoryQueue.GetLength() != 3)
		return false;
	DebugDelete(victim);
	ValidateMemoryQueue();
	if (sMemoryQueue.GetLength() != 2)
		return false;
	DebugDelete(victim5);
	ValidateMemoryQueue();
	if (sMemoryQueue.GetLength() != 1)
		return false;
	DebugDelete(victim2);
	ValidateMemoryQueue();
	if (sMemoryQueue.GetLength() != 0)
		return false;
	DebugDelete(victim4);
	return true;
}
#endif

