/*
 * Copyright (c) 1999 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * Copyright (c) 1999 Apple Computer, Inc.  All Rights Reserved.
 * The contents of this file constitute Original Code as defined in and are 
 * subject to the Apple Public Source License Version 1.1 (the "License").  
 * You may not use this file except in compliance with the License.  Please 
 * obtain a copy of the License at http://www.apple.com/publicsource and 
 * read it before using this file.
 * 
 * This Original Code and all software distributed under the License are 
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, EITHER 
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES, 
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY, FITNESS 
 * FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT.  Please see the License for 
 * the specific language governing rights and limitations under the 
 * License.
 * 
 * 
 * @APPLE_LICENSE_HEADER_END@
 */
/*
	File:		RTPSessionInterface.h

	Contains:	Implementation of object defined in .h
				
	Change History (most recent first):

	$Log: RTPSessionInterface.cpp,v $
	Revision 1.6  1999/12/21 01:44:07  lecroy
	*** empty log message ***
	
	Revision 1.5  1999/11/08 22:52:35  serenyi
	Updated code to reflect changes in QTSS API
	
	Revision 1.4  1999/10/07 08:52:32  serenyi
	Changed bool to Bool16, changed QTSS_Dictionary to QTSS_Object
	
	Revision 1.3  1999/10/06 00:05:23  murata
	support for average bit rate and removal of conditional compile flag for the new logging fields
	
	Revision 1.2  1999/09/24 19:57:27  murata
	New logging field support.
	
	Revision 1.1.1.1  1999/09/22 23:01:52  lecroy
	importing
	
	Revision 1.2  1999/09/17 02:44:44  serenyi
	Make Teardown() release the RTSPSession from the RTP session, so RTSP session can go away very quickly
	
	Revision 1.1.1.1  1999/09/13 16:16:54  rtucker
	initial checkin from rrv14
	
	
	Revision 1.2  1999/02/19 23:08:26  ds
	Created
	

*/

#include "RTPSessionInterface.h"
#include "QTSServerInterface.h"
#include "QTSS.h"
#include "OS.h"

QTSSDictionaryMap::AttrInfo	RTPSessionInterface::sAttributes[] = 
{   /*fields:   fAttrName, fFuncPtr, fIsPreemptiveSafe, fIsWriteable */
	/* 0  */ { "qtssCliSesStreamObjects", 			NULL, 	true,	false },
	/* 1  */ { "qtssCliSesCreateTimeInMsec", 		NULL, 	true,	false },
	/* 2  */ { "qtssCliSesFirstPlayTimeInMsec", 	NULL, 	true,	false },
	/* 3  */ { "qtssCliSesPlayTimeInMsec", 			NULL, 	true,	false },
	/* 4  */ { "qtssCliSesAdjustedPlayTimeInMsec",	NULL, 	true,	false },
	/* 5  */ { "qtssCliSesRTPBytesSent", 			NULL, 	true,	false },
	/* 6  */ { "qtssCliSesRTPPacketsSent", 			NULL, 	true,	false },
	/* 7  */ { "qtssCliSesState", 					NULL, 	true,	false },
	/* 8  */ { "qtssCliSesPresentationURL", 		NULL, 	true,	false },
	/* 9  */ { "qtssCliSesFirstUserAgent", 			NULL, 	true,	false },
	/* 10 */ { "qtssCliStrMovieDurationInSecs", 	NULL, 	true,	true  },
	/* 11 */ { "qtssCliStrMovieSizeInBytes", 		NULL, 	true,	true  },
	/* 12 */ { "qtssCliSesMovieAverageBitRate", 	NULL, 	true,	true  },
	/* 13 */ { "qtssCliSesLastRTSPSession", 		NULL, 	true,	false }	
};

void	RTPSessionInterface::Initialize()
{
	for (UInt32 x = 0; x < qtssCliSesNumParams; x++)
		QTSSModuleInterface::GetDictionaryMap(QTSSModuleInterface::kClientSessionDictIndex)->
			SetAttribute(x, sAttributes[x].fAttrName, sAttributes[x].fFuncPtr,
				sAttributes[x].fIsPreemptiveSafe, sAttributes[x].fIsWriteable);
}

RTPSessionInterface::RTPSessionInterface()
: 	QTSSDictionary(QTSSModuleInterface::GetDictionaryMap(QTSSModuleInterface::kClientSessionDictIndex), NULL),
	fIsFirstPlay(true),
	fFirstPlayTime(0),
	fPlayTime(0),
	fAdjustedPlayTime(0),
	fNTPPlayTime(0),
	fState(qtssPausedState),
	fPlayFlags(0),
	fSessionMutex('rsmt'),
	fRTSPSessionID(fRTSPSessionIDBuf, 0),
	fRTSPSession(NULL),
	fTimeoutTask(this, QTSServerInterface::GetServer()->GetPrefs()->GetRTPTimeoutInSecs() * 1000),
	fNumQualityLevels(0),
	fBytesSent(0),
	fPacketsSent(0),
	fMovieDuration(0),
	fMovieSizeInBytes(0),
	fMovieAverageBitRate(0)
{
	//don't actually setup the fTimeoutTask until the session has been bound!
	//(we don't want to get timeouts before the session gets bound)
	fTimeout = QTSServerInterface::GetServer()->GetPrefs()->GetRTPTimeoutInSecs() * 1000;
	
	//mark the session create time
	fSessionCreateTime = OS::Milliseconds();

	// Setup all dictionary attribute values
	
	// Make sure the dictionary knows about our preallocated memory for the RTP stream array
	this->SetEmptyVal(qtssCliSesFirstUserAgent, &fUserAgentBuffer[0], kUserAgentBufSize);
	this->SetEmptyVal(qtssCliSesStreamObjects,	 	&fStreamBuffer[0], kStreamBufSize);
	this->SetEmptyVal(qtssCliSesPresentationURL,&fPresentationURL[0], ePresentationURLSize);

	this->SetVal(qtssCliSesCreateTimeInMsec, 	&fSessionCreateTime, sizeof(fSessionCreateTime));
	this->SetVal(qtssCliSesFirstPlayTimeInMsec, &fFirstPlayTime, sizeof(fFirstPlayTime));
	this->SetVal(qtssCliSesPlayTimeInMsec, 		&fPlayTime, sizeof(fPlayTime));
	this->SetVal(qtssCliSesAdjustedPlayTimeInMsec, &fAdjustedPlayTime, sizeof(fAdjustedPlayTime));
	this->SetVal(qtssCliSesRTPBytesSent, 		&fBytesSent, sizeof(fBytesSent));
	this->SetVal(qtssCliSesRTPPacketsSent, 		&fPacketsSent, sizeof(fPacketsSent));
	this->SetVal(qtssCliSesState, 				&fState, sizeof(fState));
	this->SetVal(qtssCliSesMovieDurationInSecs, &fMovieDuration, sizeof(fMovieDuration));
	this->SetVal(qtssCliSesMovieSizeInBytes,	&fMovieSizeInBytes, sizeof(fMovieSizeInBytes));
	this->SetVal(qtssCliSesLastRTSPSession,		&fRTSPSession, sizeof(fRTSPSession));
	this->SetVal(qtssCliSesMovieAverageBitRate,	&fMovieAverageBitRate, sizeof(fMovieAverageBitRate));
}

void RTPSessionInterface::UpdateRTSPSession(RTSPSessionInterface* inNewRTSPSession)
{
	if (inNewRTSPSession != fRTSPSession)
	{
		// If there was an old session, let it know that we are done
		if (fRTSPSession != NULL)
			fRTSPSession->DecrementObjectHolderCount();
		
		// Increment this count to prevent the RTSP session from being deleted
		fRTSPSession = inNewRTSPSession;
		fRTSPSession->IncrementObjectHolderCount();
	}
}

